<?php

namespace Drupal\cms_content_sync\Commands;

use Drupal\cms_content_sync\Entity\EntityStatus;
use Drupal\cms_content_sync\Entity\Flow;
use Drupal\cms_content_sync\Entity\Pool;
use Drupal\cms_content_sync\PushIntent;
use Drupal\cms_content_sync\SyncCoreFlowExport;
use Drupal\cms_content_sync\SyncCorePoolExport;
use Drupal\cms_content_sync\SyncIntent;
use Drush\Commands\DrushCommands;
use Drush\Exceptions\UserAbortException;
use Symfony\Component\Console\Input\InputOption;

/**
 * Content Sync Drush Commands.
 */
class CMSContentSyncCommands extends DrushCommands {

  /**
   * Export the configuration to the Sync Core.
   *
   * @command cms_content_sync:configuration-export
   *
   * @aliases cse csce
   *
   * @options force
   *  Whether to ignore that another site is already using the same site ID.
   *  Useful if you change the URL of a site.
   *
   * @throws \Exception
   */
  public function configuration_export($options = ['force' => FALSE]) {
    $this->output()->writeln('Started export of pools.');
    foreach (Pool::getAll() as $pool) {
      $exporter = new SyncCorePoolExport($pool);

      if (!$options['force'] && $exporter->siteIdExists()) {
        throw new \Exception('Another site is already using the site ID "' . $pool->getSiteId() . '" for the pool "' . $pool->id . '"');
      }

      $steps = $exporter->prepareBatch();
      foreach ($steps as $step) {
        $exporter->executeBatch($step);
      }
    }
    $this->output()->writeln('Finished export of pools.');
    $this->output()->writeln('Started export of flows.');
    foreach (Flow::getAll() as $flow) {
      $exporter = new SyncCoreFlowExport($flow);
      $steps    = $exporter->prepareBatch();
      foreach ($steps as $step) {
        $exporter->executeBatch($step);
      }
    }
    $this->output()->writeln('Finished export of flows.');
  }

  /**
   * Kindly ask the Sync Core to login again.
   *
   * @command cms_content_sync:sync-core-login
   *
   * @aliases csscl
   */
  public function sync_core_login() {
    $this->output()->writeln('Calling /login for all connections.');

    $result = [];
    foreach (Flow::getAll() as $flow) {
      $exporter = new SyncCoreFlowExport($flow);
      $result = array_merge($result, $exporter->login());
    }

    foreach ($result as $url => $success) {
      if (!$success) {
        $url = preg_replace('%https?://(.*)@(.*)$%', '$2', $url);
        $this->output()->writeln('FAILED to login from ' . $url);
      }
    }

    $this->output()->writeln('Finished login.');
  }

  /**
   * Kindly ask the Sync Core to pull all entities for a specific flow.
   *
   * @command cms_content_sync:pull-entities
   *
   * @aliases cspe
   *
   * @param string $flow_id
   *   The flow the entities should be pulled from.
   * @param array $options
   *
   * @options force
   *  Also update entities which have already been pulled.
   *
   * @usage cms_content_sync:pull-entities example_flow
   *   Pulls all entities from the example flow.
   * @usage cms_content_sync:pull-entities example_flow --force
   *   Pull all entities from the "example_flow" and force entities which already have been pulled to be updated as well.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\cms_content_sync\SyncCore\Exception\SyncCoreException
   */
  public function pull_entities($flow_id, $options = ['force' => FALSE]) {
    $force = $options['force'];

    $flows = Flow::getAll();

    foreach ($flows as $id => $flow) {
      if ($flow_id && $id != $flow_id) {
        continue;
      }

      $exporter = new SyncCoreFlowExport($flow);
      $client = $exporter->getClient();

      $result = $exporter->startSync(FALSE, $force);

      if (empty($result)) {
        $this->output()->writeln('No automated pull configured for Flow: ' . $flow->label());
        continue;
      }

      $this->output()->writeln('Started pull of entities for Flow: ' . $flow->label());

      foreach ($result as $synchronization_id => $response) {
        if (!$response) {
          $this->output()->writeln('> Failed to pull! Sync ID: ' . $synchronization_id);
          continue;
        }

        if (empty($response['total'])) {
          $this->output()->writeln('> Nothing to do for: ' . $synchronization_id);
          continue;
        }

        $synchronization = $client
          ->getConnectionSynchronizationStorage()
          ->getEntity($synchronization_id);

        $runner_id = $response['id'];

        $goal = $response['total'];
        $progress = 0;

        while ($progress < $goal) {
          if ($progress > 0) {
            sleep(5);
          }

          $body = $synchronization
            ->synchronizeAllStatus($runner_id)
            ->execute()
            ->getResult();

          $progress = $body['processed'];
          if ($progress == $goal) {
            $this->output()->writeln('> Pulled ' . $goal . ' entities.');
          }
          elseif ($progress == 0) {
            sleep(5);
          }
          else {
            $this->output()->writeln('Pulled ' . $progress . ' of ' . $goal . ' entities: ' . floor($progress / $goal * 100) . '%');
          }
        }
      }
    }
  }

  /**
   * Push all entities for a specific flow.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \GuzzleHttp\Exception\GuzzleException
   *
   * @command cms_content_sync:push-entities
   *
   * @param string $flow_id
   *   The flow the entities should be pushed for.
   * @param array $options
   *
   * @options push_mode
   *  Allows to overwrite the default push mode.
   *
   * @usage cms_content_sync:push-entities example_flow
   *   Push all entities from the "example_flow" which have push configured as "automatically".
   * @usage cms_content_sync:push-entities example_flow --push_mode="automatic_manual"
   *   Push all entities from the "example_flow" which have push configured as "automatically" or "manually". Only exports manually exported entities which have not been exported before.
   * @usage cms_content_sync:push-entities example_flow --push_mode="automatic_manual_force"
   *   Push all entities from the "example_flow" which have push configured as "automatically" or "manually". Also exports entities which have not been exported before by the manual push.
   */
  public function push_entities($flow_id, $options = ['push_mode' => NULL]) {
    $push_mode = $options['push_mode'];
    $flows = Flow::getAll();

    foreach ($flows as $id => $flow) {
      if ($flow_id && $id != $flow_id) {
        continue;
      }

      /**
       * @var \Drupal\Core\Entity\EntityTypeManager $entity_type_manager
       */
      $entity_type_manager = \Drupal::service('entity_type.manager');

      foreach ($flow->getEntityTypeConfig(NULL, NULL, TRUE) as $config) {

        if ($push_mode == 'automatic_manual' || $push_mode == 'automatic_manual_force') {
          if ($config['export'] != PushIntent::PUSH_AUTOMATICALLY && $config['export'] != PushIntent::PUSH_MANUALLY) {
            continue;
          }
        }
        else {
          if ($config['export'] != PushIntent::PUSH_AUTOMATICALLY) {
            continue;
          }
        }

        $storage = $entity_type_manager->getStorage($config['entity_type_name']);

        $query = $storage
          ->getQuery();

        // Files don't have bundles, so this would lead to a fatal error then.
        if ($storage->getEntityType()->getKey('bundle')) {
          $query = $query->condition($storage->getEntityType()->getKey('bundle'), $config['bundle_name']);
        }

        $ids = $query->execute();
        $total = count($ids);
        $success = 0;
        $this->output()->writeln('Starting to push ' . $total . ' entities.');

        foreach ($ids as $id) {

          $entity = $storage->load($id);

          /**
           * @var \Drupal\cms_content_sync\Entity\EntityStatus[] $entity_status
           */
          $entity_status = EntityStatus::getInfosForEntity($entity->getEntityTypeId(), $entity->uuid(), ['flow' => $flow->id()]);

          if ($push_mode == 'automatic_manual' && (empty($entity_status) || is_null($entity_status[0]->getLastPush()))) {
            continue;
          }

          /**
           * @var \Drupal\Core\Entity\EntityTypeManager $entity_type_manager
           */
          $entity_type_manager = \Drupal::service('entity_type.manager');

          $entity = $entity_type_manager
            ->getStorage($config['entity_type_name'])
            ->load($id);

          try {
            PushIntent::pushEntity($entity, PushIntent::PUSH_FORCED, SyncIntent::ACTION_CREATE, $flow);
            $success++;
          }
          catch (\Exception $exception) {
            \Drupal::logger('cms_content_sync')->notice('Entity could not be pushed, reason: %exception', ['%exception' => $exception->getMessage()]);
          }
        }

        $this->output()->writeln('Successfully pushed ' . $success . ' entities.');

        if ($total != $success) {
          $this->output()->writeln('Failed to push ' . ($total - $success) . ' entities.');
        }
      }
    }
  }

  /**
   * Reset the status entities for a specific or all pool/s.
   *
   * @command cms_content_sync:reset-status-entities
   *
   * @aliases csrse
   *
   * @param array $options
   *
   * @options pool_id
   *  The machine name of the pool the status entities should be reset for.
   *
   * @usage cms_content_sync:reset-status-entities
   *   Reset all status entities for all pools.
   * @usage cms_content_sync:reset-status-entities --pool_id='example_pool'
   *   Reset all status entities for the "example_pool".
   *
   * @throws \Drush\Exceptions\UserAbortException.
   */
  public function reset_status_entities($options = ['pool_id' => InputOption::VALUE_OPTIONAL]) {
    $pool_id = empty($options['pool_id']) ? NULL : $options['pool_id'];
    if (empty($pool_id)) {
      $this->output()->writeln(dt('Are you sure you want to reset the status entities for all pools?'));
    }
    else {
      $this->output()->writeln(dt('Are you sure you want to reset the status entities for the pool: ' . $pool_id . '?'));
    }
    $this->output()->writeln(dt('By resetting the status of all entities, the date of the last pull and the date of the last push date will be reset. The dates will no longer be displayed until the content is pulled or pushed again and all entities will be pushed / pulled again at the next synchronization regardless of whether they have changed or not..'));

    if (!$this->io()->confirm(dt('Do you want to continue?'))) {
      throw new UserAbortException();
    }

    empty($pool_id) ? Pool::resetStatusEntities() : Pool::resetStatusEntities($pool_id);
    $this->output()->writeln('Status entities have been reset and entity caches are invalidated.');
  }

  /**
   * Check the flags for an entity.
   *
   * @command cms_content_sync:check-entity-flags
   *
   * @aliases cscef
   *
   * @param string $entity_uuid
   *   The entities uuid you would like to check for.
   * @param array $options
   *
   * @options flag The flag to check for, allowed values are: FLAG_IS_SOURCE_ENTITY, FLAG_PUSH_ENABLED, FLAG_PUSHED_AS_DEPENDENCY, FLAG_EDIT_OVERRIDE, FLAG_USER_ENABLED_PUSH, FLAG_DELETED
   *
   * @usage cms_content_sync:check-entity-flags 16cc0d54-d93d-45b8-adf2-071de9d2d32b
   *   Get all flags for the entity having the uuid = "16cc0d54-d93d-45b8-adf2-071de9d2d32b".
   * @usage cms_content_sync:check-entity-flags 16cc0d54-d93d-45b8-adf2-071de9d2d32b --flag="FLAG_EDIT_OVERRIDE"
   *   Check if the entity having the uuid = "16cc0d54-d93d-45b8-adf2-071de9d2d32b" is overridden locally.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function check_entity_flags($entity_uuid, $options = ['flag' => InputOption::VALUE_OPTIONAL]) {
    $flag = empty($options['flag']) ? NULL : $options['flag'];

    /**
     * @var \Drupal\cms_content_sync\Entity\EntityStatus[] $entity_status
     */
    $entity_status = \Drupal::entityTypeManager()->getStorage('cms_content_sync_entity_status')->loadByProperties(['entity_uuid' => $entity_uuid]);
    if (empty($entity_status)) {
      $this->output()->writeln(dt('There is no status entity existent yet for this UUID.'));
    }
    else {
      foreach ($entity_status as $status) {
        $result = '';
        $this->output()->writeln(dt('Flow: ' . $status->get('flow')->value));

        if (empty($flag)) {
          $result .= 'FLAG_IS_SOURCE_ENTITY: ' . ($status->isSourceEntity() ? 'TRUE' : 'FALSE') . PHP_EOL;
          $result .= 'FLAG_PUSH_ENABLED: ' . ($status->isPushEnabled() ? 'TRUE' : 'FALSE') . PHP_EOL;
          $result .= 'FLAG_PUSHED_AS_DEPENDENCY: ' . ($status->isPushedAsDependency() ? 'TRUE' : 'FALSE') . PHP_EOL;
          $result .= 'FLAG_EDIT_OVERRIDE: ' . ($status->isOverriddenLocally() ? 'TRUE' : 'FALSE') . PHP_EOL;
          $result .= 'FLAG_USER_ENABLED_PUSH: ' . ($status->didUserEnablePush() ? 'TRUE' : 'FALSE') . PHP_EOL;
          $result .= 'FLAG_DELETED: ' . ($status->isDeleted() ? 'TRUE' : 'FALSE') . PHP_EOL;
        }
        else {
          switch ($flag) {
            case 'FLAG_IS_SOURCE_ENTITY':
              $status->isSourceEntity() ? $result .= 'TRUE' : $result .= 'FALSE';
              break;

            case 'FLAG_PUSH_ENABLED':
              $status->isPushEnabled() ? $result .= 'TRUE' : $result .= 'FALSE';
              break;

            case 'FLAG_PUSHED_AS_DEPENDENCY':
              $status->isPushedAsDependency() ? $result .= 'TRUE' : $result .= 'FALSE';
              break;

            case 'FLAG_EDIT_OVERRIDE':
              $status->isOverriddenLocally() ? $result .= 'TRUE' : $result .= 'FALSE';
              break;

            case 'FLAG_USER_ENABLED_PUSH':
              $status->didUserEnablePush() ? $result .= 'TRUE' : $result .= 'FALSE';
              break;

            case 'FLAG_DELETED':
              $status->isDeleted() ? $result .= 'TRUE' : $result .= 'FALSE';
              break;
          }
        }
        $this->output()->writeln(dt($result));
      }
    }
  }

}
