<?php

namespace Drupal\cms_content_sync\Controller;

use Drupal\cms_content_sync\Entity\Pool;
use Drupal\cms_content_sync\SyncCore\Exception\NotFoundException;
use Drupal\cms_content_sync\SyncCore\Storage\PreviewEntityStorage;
use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Config\ConfigFactoryInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class ContentSyncSettings.
 *
 * Helper functions to set and get the settings of this page.
 *
 * @package Drupal\cms_content_sync
 */
class ContentSyncSettings {
  /**
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * Singleton.
   *
   * @return ContentSyncSettings
   */
  public static function getInstance() {
    static $instance = NULL;
    if ($instance) {
      return $instance;
    }

    return $instance = self::createInstance(\Drupal::getContainer());
  }

  /**
   * Create a new instance of the controller with the services of the given container.
   *
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   *
   * @return ContentSyncSettings
   */
  public static function createInstance(ContainerInterface $container) {
    return new static(
      $container->get('config.factory')
    );
  }

  /**
   * Constructs a new EntityListBuilder object.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   */
  public function __construct(ConfigFactoryInterface $configFactory) {
    $this->configFactory = $configFactory;
  }

  /**
   * @var bool
   */
  protected $isPreviewEnabled = NULL;

  /**
   * @return bool
   */
  public function isPreviewEnabled() {
    if ($this->isPreviewEnabled !== NULL) {
      return $this->isPreviewEnabled;
    }

    return $this->isPreviewEnabled = boolval(
      $this
        ->configFactory
        ->get('cms_content_sync.settings')
        ->get('cms_content_sync_enable_preview')
    );
  }

  /**
   * @param bool $set
   */
  public function setPreviewEnabled($set) {
    $this
      ->configFactory
      ->getEditable('cms_content_sync.settings')
      ->set('cms_content_sync_enable_preview', $set)
      ->save();

    $this->isPreviewEnabled = $set;
  }

  /**
   * @var string
   */
  protected $siteBaseUrl = NULL;

  /**
   * @param bool $config_only
   *   Whether to return the base URL as per the settings only. Otherwise it will return the
   *   default site URL.
   * @return string
   * @throws \Exception
   */
  public function getSiteBaseUrl($config_only = FALSE) {
    if ($this->siteBaseUrl !== NULL) {
      return $this->siteBaseUrl;
    }

    // Check if the base_url is overwritten within the settings.
    $setting = $this
      ->configFactory
      ->get('cms_content_sync.settings')
      ->get('cms_content_sync_base_url');

    if ($config_only) {
      return $setting;
    }

    if ($setting) {
      // Validate the Base URL.
      if (UrlHelper::isValid($setting, TRUE) && mb_substr($setting, -1) !== '/') {
        return $this->siteBaseUrl = $setting;
      }
      else {
        throw new \Exception(t('The configured base URL is not a valid URL. Ensure that it does not contain a trailing slash.'));
      }
    }

    global $base_url;
    return $this->siteBaseUrl = $base_url;
  }

  /**
   * @param string $set
   */
  public function setSiteBaseUrl($set) {
    $this
      ->configFactory
      ->getEditable('cms_content_sync.settings')
      ->set('cms_content_sync_base_url', $set)
      ->save();

    // Reset so it's computed again.
    $this->siteBaseUrl = NULL;
  }

  /**
   * Whether or not users can directly communicate with the Sync Core. This only makes sense if the Sync Core is
   * exposed publicly so right now we restricted it to the Cloud version of Content Sync.
   * The setting is saved in the Sync Core, not locally at this site. Otherwise the configuration of multiple sites
   * would conflict and lead to undesired outcomes.
   *
   * @param bool $default
   *   What to assume as the default value if no explicit settings are saved yet.
   *
   * @return bool
   *
   * @throws \Drupal\cms_content_sync\SyncCore\Exception\SyncCoreException
   */
  public function isDirectSyncCoreAccessEnabled($default = TRUE) {
    static $cache = NULL;

    if ($cache !== NULL) {
      return $cache;
    }

    $checked = [];

    // We don't distinguish between multiple Sync Cores. If it's enabled for one, it's enabled for all.
    foreach (Pool::getAll() as $pool) {
      $url = $pool->getBackendUrl();
      if (in_array($url, $checked)) {
        continue;
      }
      $checked[] = $url;

      try {
        $connection = $pool
          ->getClient()
          ->getConnectionStorage()
          ->getItem(PreviewEntityStorage::ID)
          ->execute()
          ->getItem();
      }
      catch (NotFoundException $e) {
        // No config exported yet, so skip this.
        continue;
      }

      if ($connection) {
        if (isset($connection['options'][PreviewEntityStorage::PUBLIC_ACCESS_OPTION_NAME])) {
          if ($connection['options'][PreviewEntityStorage::PUBLIC_ACCESS_OPTION_NAME]) {
            return $cache = TRUE;
          }
          else {
            $cache = FALSE;
          }
        }
      }
    }

    // If it's not set at all, we default it to TRUE so people benefit from the increased performance.
    if ($cache === NULL) {
      return $cache = $default;
    }

    return $cache = FALSE;
  }

  /**
   * Set whether or not to allow direct Sync Core communication.
   *
   * @param bool $set
   *
   * @throws \Exception
   */
  public function setDirectSyncCoreAccessEnabled($set) {
    $checked = [];

    // We don't distinguish between multiple Sync Cores. If it's enabled for one, it's enabled for all.
    foreach (Pool::getAll() as $pool) {
      $url = $pool->getBackendUrl();
      if (in_array($url, $checked)) {
        continue;
      }
      $checked[] = $url;

      try {
        /**
         * @var \Drupal\cms_content_sync\SyncCore\Storage\ConnectionStorage $storage
         */
        $storage = $pool
          ->getClient()
          ->getConnectionStorage();

        /**
         * @var \Drupal\cms_content_sync\SyncCore\Entity\EntityPreviewConnection $connection
         */
        $connection = $storage
          ->getEntity(PreviewEntityStorage::ID);

        $connection
          ->allowPublicAccess($set)
          ->execute();
      }
      catch (NotFoundException $e) {
        // No config exported yet, so skip this.
        continue;
      }
    }
  }

  /**
   * @var string
   */
  protected $siteId = NULL;

  /**
   * @return bool
   */
  public function getSiteId() {
    if ($this->siteId !== NULL) {
      return $this->siteId;
    }

    foreach (Pool::getAll() as $pool) {
      $this->siteId = $pool->getSiteId();
      break;
    }

    return $this->siteId;
  }

}
