<?php

namespace Drupal\cms_content_sync\Controller;

use Drupal\cms_content_sync\SyncCore;
use Drupal\Core\Url;
use Drupal\Core\Controller\ControllerBase;
use Drupal\cms_content_sync\SyncCoreFlowExport;

/**
 * Pull controller.
 */
class FlowPull extends ControllerBase {

  /**
   * Pull all entities of this Flow.
   *
   * @param $cms_content_sync_flow
   * @param $pull_mode
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse|null
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function pull($cms_content_sync_flow, $pull_mode) {
    /**
     * @var \Drupal\cms_content_sync\Entity\Flow $flow
     */
    $flow = \Drupal::entityTypeManager()
      ->getStorage('cms_content_sync_flow')
      ->load($cms_content_sync_flow);

    $force_pull = FALSE;
    if ($pull_mode == 'all_entities') {
      $force_pull = TRUE;
    }

    $handler = new SyncCoreFlowExport($flow);
    $result = $handler->startSync(FALSE, $force_pull);

    $steps = [];

    foreach ($result as $id => $response) {
      if (!$response) {
        $steps[] = [
          'type' => 'FAILURE',
        ];
        continue;
      }

      if (empty($response['total'])) {
        $steps[] = [
          'type' => 'EMPTY',
        ];
        continue;
      }

      for ($i = 1; $i <= $response['total']; $i++) {
        $steps[] = [
          'type' => 'REQUEST',
          'goal' => $i,
          'total' => $response['total'],
          'sync_core_url' => $handler->getClient()->getBaseUrl(),
          'synchronization_id' => $id,
          'runner_id' => $response['id'],
        ];
      }
    }

    $operations = [];
    foreach ($steps as $step) {
      $operations[] = [
        '\Drupal\cms_content_sync\Controller\FlowPull::batch',
        [$step],
      ];
    }

    $batch = [
      'title' => t('Pull all'),
      'operations' => $operations,
      'finished' => '\Drupal\cms_content_sync\Controller\FlowPull::batchFinished',
    ];
    batch_set($batch);

    return batch_process(Url::fromRoute('entity.cms_content_sync_flow.collection'));
  }

  /**
   * Batch pull finished callback.
   *
   * @param $success
   * @param $results
   * @param $operations
   */
  public static function batchFinished($success, $results, $operations) {
    if ($success) {
      $message = t('Flow has been pulled.');
    }
    else {
      $message = t('Flow pull failed.');
    }

    \Drupal::messenger()->addMessage($message);

    $failed = 0;
    $empty = 0;
    $synchronized = 0;
    foreach ($results as $result) {
      if ($result['type'] == 'FAILURE') {
        $failed++;
      }
      elseif ($result['type'] == 'EMPTY') {
        $empty++;
      }
      else {
        $synchronized += $result['total'];
      }
    }

    if ($failed) {
      \Drupal::messenger()->addMessage(t('Failed to pull from %failed entity pools.', ['%failed' => $failed]));
    }
    if ($empty) {
      \Drupal::messenger()->addMessage(t('%empty entity pools were empty or had no new entities.', ['%empty' => $empty]));
    }
    if ($synchronized) {
      \Drupal::messenger()->addMessage(t('%synchronized entities have been pulled.', ['%synchronized' => $synchronized]));
    }
  }

  /**
   * Batch pull callback for the pull-all operation.
   *
   * @param $operation
   * @param $context
   */
  public static function batch($operation, &$context) {
    $message = 'Pulling...';
    $results = [];
    if (isset($context['results'])) {
      $results = $context['results'];
    }

    if ($operation['type'] == 'FAILURE' || $operation['type'] == 'EMPTY') {
      $results[] = $operation;
    }
    else {
      $sync_core_url = $operation['sync_core_url'];
      $synchronization_id = $operation['synchronization_id'];
      $runner_id = $operation['runner_id'];

      $client = SyncCore::getClient($sync_core_url);

      if (!isset($context['progress'][$synchronization_id])) {
        $context['progress'][$synchronization_id] = 0;
      }

      $goal = $operation['goal'];
      $total = $operation['total'];

      while ($context['progress'][$synchronization_id] < $goal) {
        if ($context['progress'][$synchronization_id] > 0) {
          sleep(5);
        }

        $context['progress'][$synchronization_id] = $client
          ->getConnectionSynchronizationStorage()
          ->getEntity($synchronization_id)
          ->synchronizeAllStatus($runner_id)
          ->execute()
          ->getResult()['processed'];
      }

      if ($context['progress'][$synchronization_id] == $goal && $goal == $total) {
        $results[] = ['type' => 'success', 'total' => $total];
      }
    }

    $context['message'] = $message;
    $context['results'] = $results;
  }

}
