<?php

namespace Drupal\cms_content_sync\Controller;

use Drupal\Component\Render\FormattableMarkup;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\Entity\ConfigEntityListBuilder;
use Drupal\Core\Entity\EntityInterface;
use Drupal\cms_content_sync\Form\PoolForm;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a listing of Pool.
 */
class PoolListBuilder extends ConfigEntityListBuilder {

  /**
   * Constructs a new EntityListBuilder object.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type definition.
   * @param \Drupal\Core\Entity\EntityStorageInterface $storage
   *   The entity storage class.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   */
  public function __construct(EntityTypeInterface $entity_type, EntityStorageInterface $storage, ConfigFactoryInterface $config_factory) {
    parent::__construct($entity_type, $storage);
    $this->configFactory = $config_factory;
  }

  /**
   * {@inheritdoc}
   */
  public static function createInstance(ContainerInterface $container, EntityTypeInterface $entity_type) {
    return new static(
      $entity_type,
      $container->get('entity.manager')->getStorage($entity_type->id()),
      $container->get('config.factory')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildHeader() {
    $header['name'] = $this->t('Name');
    $header['id'] = $this->t('Machine name');
    $header['site_id'] = $this->t('Site identifier');
    $header['backend_url'] = $this->t('Sync Core URL');
    return $header + parent::buildHeader();
  }

  /**
   * {@inheritdoc}
   */
  public function buildRow(EntityInterface $entity) {
    /**
     * @var \Drupal\cms_content_sync\Entity\Pool $entity
     */
    $row['name'] = $entity->label();
    $row['id'] = $entity->id();
    $row['site_id'] = $entity->getSiteId();
    $row['backend_url'] = $entity->getBackendUrl();

    // Check of overwrites.
    $config = $this->configFactory->get('cms_content_sync.pool.' . $entity->id());

    $overwritten_site_id = $config->get('site_id') != $entity->getSiteId();
    $overwritten_backend_url = $config->get('backend_url') != $entity->getBackendUrl();

    if ($overwritten_site_id) {
      $row['site_id'] .= ' <i>(' . $this->t('Overwritten') . ')</i>';
    }

    if ($overwritten_backend_url) {
      $row['backend_url'] .= ' <i>(' . $this->t('Overwritten') . ')</i>';
    }

    if (strlen($entity->getSiteId()) > PoolForm::siteIdMaxLength) {
      $messenger = \Drupal::messenger();
      $warning = 'The site id of pool ' . $entity->id() . ' is having more then ' . PoolForm::siteIdMaxLength . ' characters. This is not allowed due to backend limitations and will result in an exception when it is trying to be exported.';
      $messenger->addWarning(t($warning));
    }

    $row['site_id'] = new FormattableMarkup($row['site_id'], []);
    $row['backend_url'] = new FormattableMarkup($row['backend_url'], []);

    return $row + parent::buildRow($entity);
  }

  /**
   *
   */
  public function render() {
    $build = parent::render();
    $overwritten = FALSE;
    $rows = $build["table"]["#rows"];

    if (!empty($rows)) {
      foreach ($rows as $row) {
        if (strpos($row['site_id']->__toString(), 'Overwritten') || strpos($row['backend_url']->__toString(), 'Overwritten')) {
          $overwritten = TRUE;
        }
      }

      if ($overwritten) {
        $build['explanation_overwritten'] = [
          '#markup' => '<i>' . $this->t('Overwritten: This configuration has been overwritten within your settings.php file.') . '</i><br>',
        ];
      }

      $build['hints'] = [
        '#markup' => '<h3>' . $this->t('Hints') . '</h3>' . $this->t('You can overwrite parts of your pool configuration using your settings.php file:'),
      ];

      $overwrite_site_id_example = '<li>' . $this->t('Site ID: <i>@overwrite_site_id</i>', ['@overwrite_site_id' => '$settings["cms_content_sync"]["pools"][<pool_machine_name>]["site_id"] = "example";']) . '</li>';
      $overwrite_sync_core_url_example = '<li>' . $this->t('Sync Core URL: <i>@overwrite_sync_core_url</i>', ['@overwrite_sync_core_url' => '$settings["cms_content_sync"]["pools"][<pool_machine_name>]["backend_url"] = "http://example.com";']) . '</li>';

      $build['hints_examples'] = [
        '#markup' => '<ul>' . $overwrite_site_id_example . $overwrite_sync_core_url_example . '</ul>',
      ];
    }

    return $build;
  }

}
