<?php

namespace Drupal\cms_content_sync\Form;

use Drupal\cms_content_sync\Controller\AuthenticationByUser;
use Drupal\cms_content_sync\Controller\ContentSyncSettings;
use Drupal\cms_content_sync\SyncCore;
use Drupal\cms_content_sync\SyncCore\Exception\BadRequestException;
use Drupal\cms_content_sync\SyncCore\Exception\ForbiddenException;
use Drupal\cms_content_sync\SyncCore\Exception\NotFoundException;
use Drupal\cms_content_sync\SyncCore\Exception\SyncCoreException;
use Drupal\cms_content_sync\SyncCore\Helper;
use Drupal\cms_content_sync\SyncCore\Query\Condition\DataCondition;
use Drupal\cms_content_sync\SyncCore\Query\Condition\ParentCondition;
use Drupal\Core\Link;
use Drupal\cms_content_sync\Entity\Flow;
use Drupal\cms_content_sync\Entity\Pool;
use Drupal\cms_content_sync\SyncCore\Client;
use Drupal\cms_content_sync\SyncCore\Storage\ApiStorage;
use Drupal\cms_content_sync\SyncCore\Storage\InstanceStorage;
use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandler;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Site\Settings;

/**
 * Form handler for the Pool add and edit forms.
 */
class PoolForm extends EntityForm {

  const AUTHENTICATION_TYPE_DESCRIPTION = 'Define how the Sync Core should authenticate against this page. Using Basic Auth can improve startup times when connecting a lot of sites and resolves issues when hiding the default /user/login route for security reasons. When using the SaaS product, we recommend using the standard Drupal login. To enable Basic Auth, please enable the basic_auth module.';

  /**
   * @var int Defines the max length for the siteID. This must be limited due to the maximum characters allowed for table names within mongo db.
   */
  const siteIdMaxLength = 20;

  /**
   * Constructs an PoolForm object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entityTypeManager.
   * @param \Drupal\Core\Extension\ModuleHandler $moduleHandler
   *   The moduleHandler.
   */
  public function __construct(EntityTypeManagerInterface $entityTypeManager, ModuleHandler $moduleHandler) {
    $this->entityTypeManager = $entityTypeManager;
    $this->moduleHandler = $moduleHandler;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('module_handler')
    );
  }

  /**
   * @var string STEP_SYNC_CORE Select a Sync Core another pool on this site already uses or enter a new one.
   */
  const STEP_SYNC_CORE = 'sync-core';

  /**
   * @var string STEP_SITE_ID Select the ID for this site or enter a new one.
   */
  const STEP_SITE_ID = 'site-id';

  /**
   * @var string STEP_POOL Select pool from existing pools in the Sync Core or enter a new one.
   */
  const STEP_POOL = 'pool';

  /**
   * @var string CONTAINER_ID The element ID of the container that's replaced with every AJAX request.
   */
  const CONTAINER_ID = 'pool-form-container';

  /**
   * Get default values that were provided in the URL.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @return array
   */
  protected function getDefaults(FormStateInterface $form_state) {
    return [
      'backend_url' => empty($_GET['backend_url']) ? $form_state->getValue('default_backend_url') : $_GET['backend_url'],
      'name' => empty($_GET['name']) ? $form_state->getValue('default_name') : $_GET['name'],
      'id' => empty($_GET['id']) ? $form_state->getValue('default_id') : $_GET['id'],
    ];
  }

  /**
   * Return the current step of our multi-step form.
   *
   * @return string
   */
  protected function getCurrentFormStep(FormStateInterface $form_state) {
    /**
     * @var \Drupal\cms_content_sync\Entity\Pool $pool
     */
    $pool = $this->entity;

    if (!$pool->getBackendUrl() && !$form_state->getValue('backend_url')) {
      return self::STEP_SYNC_CORE;
    }

    if (!$pool->getSiteId() && !$form_state->getValue('site_id')) {
      return self::STEP_SITE_ID;
    }

    return self::STEP_POOL;
  }

  /**
   * Return the current step of our multi-step form.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @return string
   */
  protected function getLastFormStep(FormStateInterface $form_state) {
    $step = $form_state->getValue('step');
    if (empty($step)) {
      return self::STEP_SYNC_CORE;
    }

    return $step;
  }

  /**
   * Return next form.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @return array The bundle settings.
   */
  public function ajaxReturn($form, FormStateInterface $form_state) {
    return $form['elements'];
  }

  /**
   * Rebuild form for next step.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public function connectSyncCore($form, FormStateInterface $form_state) {
    $form_state->setRebuild();
  }

  /**
   * Rebuild form for next step.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public function checkSiteId($form, FormStateInterface $form_state) {
    $form_state->setRebuild();
  }

  /**
   * Rebuild form for next step.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public function createNew($form, FormStateInterface $form_state) {
    $form_state->setValue('id', 'new');

    // If we only use ->setValue() the default value in the input element will still be whatever was selected at the radios.
    $data = $form_state->getUserInput();
    $data['id'] = 'new';
    $form_state->setUserInput($data);

    $form_state->setRebuild();
  }

  /**
   * Step 2: Enter or select site ID.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   * @param bool $collapsed
   *
   * @return array the form
   */
  protected function syncCoreForm(array $form, FormStateInterface $form_state, $collapsed = FALSE) {
    /**
     * @var \Drupal\cms_content_sync\Entity\Pool $pool
     */
    $pool = $this->entity;

    if ($collapsed) {
      $elements['backend_url'] = [
        '#type' => 'hidden',
        '#value' => $form_state->getValue('backend_url') ? $form_state->getValue('backend_url') : $pool->getBackendUrl(),
      ];

      $elements['authentication_type'] = [
        '#type' => 'hidden',
        '#value' => $form_state->getValue('authentication_type') ? $form_state->getValue('authentication_type') : $pool->getAuthenticationType(),
      ];

      return $elements;
    }

    $elements['headline'] = [
      '#markup' => '<br><br><h1>Step 1: Pick Sync Core</h1>' .
      '<div class="messages messages--status">' .
      $this->t("The Sync Core is the central distribution service used to syndicate content between sites. If you don't have one yet, just <a href='mailto:info@cms-content-sync.io?subject=Evaluation%20Sync%20Core&body=Please%20provide%20me%20an%20evaluation%20Sync%20Core'>contact us</a>. <br>
<br>
<strong>Want to learn more?</strong><br>
Use <a href='https://www.cms-content-sync.io/content-staging/'>Content Staging</a> to publish content from a private site to a public site. Pricing for Content Staging is documented <a href='https://www.cms-content-sync.io/content-sync-for-drupal/content-staging-for-drupal/'>here</a>.<br>
Use <a href='https://www.cms-content-sync.io/content-syndication/'>Content Syndication</a> to connect many public sites to share content between them. Pricing for Content Syndication is documented <a href='https://www.cms-content-sync.io/content-sync-for-drupal/content-syndication-for-drupal/'>here</a>.
") .
      '</div><br><br>',
    ];

    $default_backend_url = $this->getDefaults($form_state)['backend_url'];

    $pools = Pool::getAll();
    $urls = [];
    $default_auth = NULL;
    foreach ($pools as $existing_pool) {
      $url = $existing_pool->getBackendUrl();
      $default_auth = $existing_pool->getAuthenticationType();
      $urls[$url] = Helper::obfuscateCredentials($url);
    }

    if (count($urls) && !isset($_GET['add-sync-core']) && empty($default_backend_url)) {
      $url = $pool->getBackendUrl();

      if ($url) {
        $urls[$url] = Helper::obfuscateCredentials($url);
      }

      $elements['backend_url'] = [
        '#type' => 'radios',
        '#title' => $this->t('Sync Core URL'),
        '#default_value' => $url ? $url : array_slice(array_keys($urls), 0, 1)[0],
        '#description' => $this->t("The backend url can be overwritten within your environment specific settings.php file by using <i>@settings</i>.", [
          '@settings' => '$settings["cms_content_sync"]["pools"]["' . $this->configMachineName . '"]["backend_url"] = "http://cms-content-sync-example:8691/rest"',
          '@config_machine_name' => $this->configMachineName,
        ]) . '<br><a href="?add-sync-core">Add new</a>',
        '#options' => $urls,
        '#required' => TRUE,
      ];
    }
    else {
      $elements['backend_url'] = [
        '#type' => 'url',
        '#title' => $this->t('Sync Core URL'),
        '#default_value' => empty($default_backend_url) ? $pool->getBackendUrl() : $default_backend_url,
        '#description' => $this->t("The backend url can be overwritten within your environment specific settings.php file by using <i>@settings</i>.", [
          '@settings' => '$settings["cms_content_sync"]["pools"]["' . $this->configMachineName . '"]["backend_url"] = "http://cms-content-sync-example.de:8691/rest"',
          '@config_machine_name' => $this->configMachineName,
        ]),
        '#required' => TRUE,
      ];
    }

    // If the backend_url is set within the settings.php,
    // the form field is disabled.
    if (isset($this->backendUrl)) {
      $elements['backend_url']['#disabled'] = TRUE;
      $elements['backend_url']['#default_value'] = $this->backendUrl;
      $elements['backend_url']['#description'] = $this->t('The backend url is set within the environment specific settings.php file.');
    }

    $auth_options = [
      Pool::AUTHENTICATION_TYPE_COOKIE => $this->t("Standard (Cookie)"),
    ];
    if ($this->moduleHandler->moduleExists('basic_auth')) {
      $auth_options[Pool::AUTHENTICATION_TYPE_BASIC_AUTH] = $this->t("Basic Auth");
    }

    $elements['authentication_type'] = [
      '#type' => 'select',
      '#title' => $this->t('Authentication'),
      '#default_value' => $pool->getAuthenticationType() ? $pool->getAuthenticationType() :
      ($default_auth ? $default_auth :
          (isset($auth_options[Pool::AUTHENTICATION_TYPE_BASIC_AUTH]) ? Pool::AUTHENTICATION_TYPE_BASIC_AUTH :
            Pool::AUTHENTICATION_TYPE_COOKIE
          )
      ),
      '#description' => $this->t(PoolForm::AUTHENTICATION_TYPE_DESCRIPTION),
      '#options' => $auth_options,
      '#required' => TRUE,
    ];

    $elements['continue'] = [
      '#prefix' => '<br><br>',
      '#type' => 'submit',
      '#submit' => ['::connectSyncCore'],
      '#value' => $this->t("Connect"),
      '#name' => 'connect',
      // '#limit_validation_errors' => [],
      '#attributes' => [
        'class' => ['button--primary'],
      ],
      '#ajax' => [
        'callback' => '::ajaxReturn',
        'wrapper' => self::CONTAINER_ID,
        'method' => 'replace',
        'effect' => 'fade',
        'progress' => [
          'type' => 'throbber',
          'message' => "loading settings...",
        ],
      ],
    ];

    return $elements;
  }

  /**
   * Step 1: Enter or select Sync Core URL.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @return array the form
   */
  protected function siteIdForm(array $form, FormStateInterface $form_state, $collapsed = FALSE) {
    $elements = $this->syncCoreForm($form, $form_state, TRUE);

    /**
     * @var \Drupal\cms_content_sync\Entity\Pool $pool
     */
    $pool = $this->entity;

    if ($collapsed) {
      $elements['site_id'] = [
        '#type' => 'hidden',
        '#value' => $form_state->getValue('site_id') ? $form_state->getValue('site_id') : $pool->getSiteId(),
      ];

      return $elements;
    }

    $elements['headline'] = [
      '#markup' => '<br><br><h1>Step 2: ID your site</h1>',
    ];

    $pools = Pool::getAll();
    $ids = [];
    foreach ($pools as $existing_pool) {
      $site_id = $existing_pool->getSiteId();
      $ids[$site_id] = $site_id;
    }

    if (count($ids) && !isset($_GET['change-site-id'])) {
      $site_id = $pool->getSiteId();

      if ($site_id) {
        $ids[$site_id] = $site_id;
      }

      $elements['site_id'] = [
        '#type' => 'radios',
        '#title' => $this->t('Site identifier'),
        '#default_value' => $site_id ? $site_id : reset($ids),
        '#description' => $this->t("This identifier will be used to identify the origin of entities on other sites and is used as a machine name for identification. Due to backend limitations, it can not have more then @max_length characters.
      Once connected, you cannot change this identifier. Typically you want to use the fully qualified domain name of this website as an identifier.<br>
      The Site identifier can be overwritten within your environment specific settings.php file by using <i>@settings</i>.<br>
      If you do so, you should exclude the Site identifier for this configuration from the configuration import/export by using the module <a href='https://www.drupal.org/project/config_ignore' target='_blank'>Config ignore</a>.
      The exclude could for example look like this: <i>cms_content_sync.pool.@config_machine_name:site_id</i><br>
      <i>Hint: If this configuration is saved before the value with the settings.php got set, you need to re-save this configuration once the value within the settings.php got set.</i>", [
        '@settings' => '$settings["cms_content_sync"]["pools"]["' . $this->configMachineName . '"]["site_id"] = "my-site-identifier"',
        '@config_machine_name' => $this->configMachineName,
        '@max_length' => PoolForm::siteIdMaxLength,
      ]),
        '#required' => TRUE,
        '#options' => $ids,
      ];
    }
    else {
      $elements['site_id'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Site identifier'),
        '#default_value' => $pool->getSiteId(),
        '#description' => $this->t("This identifier will be used to identify the origin of entities on other sites and is used as a machine name for identification. Due to backend limitations, it can not have more then @max_length characters.
      Once connected, you cannot change this identifier. Typically you want to use the fully qualified domain name of this website as an identifier.<br>
      The Site identifier can be overwritten within your environment specific settings.php file by using <i>@settings</i>.<br>
      If you do so, you should exclude the Site identifier for this configuration from the configuration import/export by using the module <a href='https://www.drupal.org/project/config_ignore' target='_blank'>Config ignore</a>.
      The exclude could for example look like this: <i>cms_content_sync.pool.@config_machine_name:site_id</i><br>
      <i>Hint: If this configuration is saved before the value with the settings.php got set, you need to re-save this configuration once the value within the settings.php got set.</i>", [
        '@settings' => '$settings["cms_content_sync"]["pools"]["' . $this->configMachineName . '"]["site_id"] = "my-site-identifier"',
        '@config_machine_name' => $this->configMachineName,
        '@max_length' => PoolForm::siteIdMaxLength,
      ]),
        '#required' => TRUE,
        '#maxlength' => PoolForm::siteIdMaxLength,
      ];
    }

    // If the site id is set within the settings.php,
    // the form field is disabled.
    if (isset($this->overwrittenSiteId)) {
      $elements['site_id']['#disabled'] = TRUE;
      $elements['site_id']['#default_value'] = $this->overwrittenSiteId;
      $elements['site_id']['#description'] = $this->t('Site identifier is set within the environment specific settings.php file. The value may automatically be trimmed to @max_length characters due to backend limitations', [
        '@max_length' => PoolForm::siteIdMaxLength,
      ]);
    }

    $elements['continue'] = [
      '#prefix' => '<br><br>',
      '#type' => 'submit',
      '#submit' => ['::checkSiteId'],
      '#value' => $this->t("Continue"),
      '#name' => 'validate',
      // '#limit_validation_errors' => [],
      '#attributes' => [
        'class' => ['button--primary'],
      ],
      '#ajax' => [
        'callback' => '::ajaxReturn',
        'wrapper' => self::CONTAINER_ID,
        'method' => 'replace',
        'effect' => 'fade',
        'progress' => [
          'type' => 'throbber',
          'message' => "loading settings...",
        ],
      ],
    ];

    return $elements;
  }

  /**
   * List all remote pools that aren't used locally yet.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @throws \Exception
   */
  protected function getRemotePools(FormStateInterface $form_state) {
    $client = SyncCore::getClient($form_state->getValue('backend_url'));
    $api_storage = new ApiStorage($client);
    $remote_pools = $api_storage
      ->listItems()
      ->setCondition(
        ParentCondition::all()
          ->add(DataCondition::equal(ApiStorage::PROPERTY_VERSION, ApiStorage::CUSTOM_API_VERSION))
          ->add(DataCondition::equal(ApiStorage::PROPERTY_PARENT_ID, 'drupal-' . ApiStorage::CUSTOM_API_VERSION))
      )
      ->execute()
      ->getAll();

    $pools = Pool::getAll();

    $options = [];
    foreach ($remote_pools as $option) {
      $id = preg_replace('@-' . ApiStorage::CUSTOM_API_VERSION . '$@', '', $option['id']);

      // Already exists locally.
      if (isset($pools[$id])) {
        continue;
      }

      $options[$id] = $option['name'];
    }

    return $options;
  }

  /**
   * Step 3: Select an existing pool or create a new one.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @return array the form
   *
   * @throws \Exception
   */
  protected function poolForm(array $form, FormStateInterface $form_state) {
    $elements = $this->siteIdForm($form, $form_state, TRUE);

    $elements['headline'] = [
      '#markup' => '<br><br><h1>Step 3: Pool properties</h1>',
    ];

    /**
     * @var \Drupal\cms_content_sync\Entity\Pool $pool
     */
    $pool = $this->entity;

    $default_name = $this->getDefaults($form_state)['name'];
    $default_id = $this->getDefaults($form_state)['id'];

    $options = $pool->isNew() && empty($default_id) ? $this->getRemotePools($form_state) : [];

    if (count($options) && !$form_state->getValue('id')) {
      $elements['id'] = [
        '#type' => 'radios',
        '#title' => $this->t('Existing pools'),
        '#required' => TRUE,
        '#default_value' => array_slice(array_keys($options), 0, 1)[0],
        '#options' => $options,
      ];
    }
    else {
      $elements['label'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Label'),
        '#maxlength' => 255,
        '#default_value' => empty($default_name) ? ($form_state->getValue('label') ? $form_state->getValue('label') :
        ($pool->label() ? $pool->label() : (empty(Pool::getAll()['content']) ? $this->t('Content') : ''))) : $default_name,
        '#description' => $this->t("The pool name. If you aren't working with multiple pools just leave it as it is."),
        '#required' => TRUE,
      ];

      $elements['id'] = [
        '#type' => 'machine_name',
        '#default_value' => empty($default_id) ? ($form_state->getValue('id') ? $form_state->getValue('id') :
        ($pool->id() ? $pool->id() : (empty(Pool::getAll()['content']) ? 'content' : ''))) : $default_id,
        '#machine_name' => [
          'exists' => [$this, 'exist'],
        ],
        '#description' => $this->t("A unique ID that must be identical on all sites that want to connect to this pool. If you aren't working with multiple pools just leave it as it is."),
        '#disabled' => !$pool->isNew(),
      ];
    }

    // AJAX? Show submit buttons inline.
    if ($form_state->getTriggeringElement()) {
      $actions = $this->actions($form, $form_state);
      $actions['submit']['#attributes']['class'][] = 'button--primary';
      $elements = array_merge($elements, $actions);
    }

    if (!isset($elements['label'])) {
      $elements['create'] = [
        '#prefix' => '<br><br>',
        '#type' => 'submit',
        '#submit' => ['::createNew'],
        '#value' => $this->t("Create new"),
        '#name' => 'create',
        '#ajax' => [
          'callback' => '::ajaxReturn',
          'wrapper' => self::CONTAINER_ID,
          'method' => 'replace',
          'effect' => 'fade',
          'progress' => [
            'type' => 'throbber',
            'message' => "loading settings...",
          ],
        ],
      ];
    }

    return $elements;
  }

  /**
   * @var string|null
   */
  protected $backendUrl = NULL;

  /**
   * @var string|null
   */
  protected $configMachineName = NULL;

  /**
   * @var string|null
   */
  protected $overwrittenSiteId = NULL;

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);

    $form['#tree'] = FALSE;

    $defaults = $this->getDefaults($form_state);
    if (!empty($defaults['backend_url'])) {
      $form['default_backend_url'] = [
        '#type' => 'hidden',
        '#value' => $defaults['backend_url'],
      ];
    }
    if (!empty($defaults['name'])) {
      $form['default_name'] = [
        '#type' => 'hidden',
        '#value' => $defaults['name'],
      ];
    }
    if (!empty($defaults['id'])) {
      $form['default_id'] = [
        '#type' => 'hidden',
        '#value' => $defaults['id'],
      ];
    }

    /**
     * @var \Drupal\cms_content_sync\Entity\Pool $pool
     */
    $pool = $this->entity;

    // Check if the site id or backend_url got set within the settings*.php.
    if (!is_null($pool->id)) {
      $this->configMachineName = $pool->id;
      $cms_content_sync_settings = Settings::get('cms_content_sync');
      if (!is_null($cms_content_sync_settings) && isset($cms_content_sync_settings['pools'][$pool->id]['site_id'])) {
        // When it is set, we anyway need to ensure that it is not having more then PoolForm::siteIdMaxLength characters.
        $this->overwrittenSiteId = $cms_content_sync_settings['pools'][$pool->id]['site_id'];
        if (strlen($this->overwrittenSiteId) > PoolForm::siteIdMaxLength) {
          $messenger = \Drupal::messenger();
          $warning = 'The site id is having more then ' . PoolForm::siteIdMaxLength . ' characters. This is not allowed due to backend limitations and will result in an exception when it is trying to be exported.';
          $messenger->addWarning(t($warning));
        }
      }
      if (!is_null($cms_content_sync_settings) && isset($cms_content_sync_settings['pools'][$pool->id]['backend_url'])) {
        $this->backendUrl = $cms_content_sync_settings['pools'][$pool->id]['backend_url'];
      }
    }
    if (!isset($this->configMachineName)) {
      $this->configMachineName = '<machine_name_of_the_configuration>';
    }

    $step = $this->getCurrentFormStep($form_state);

    if ($step === self::STEP_SYNC_CORE) {
      $elements = $this->syncCoreForm($form, $form_state);
    }
    elseif ($step === self::STEP_SITE_ID) {
      $elements = $this->siteIdForm($form, $form_state);
    }
    else {
      $elements = $this->poolForm($form, $form_state);
    }

    $form['elements'] = array_merge([
      '#prefix' => '<div id="' . self::CONTAINER_ID . '">',
      '#suffix' => '</div>',
      'step' => [
        '#type' => 'hidden',
        '#value' => $step,
      ],
    ], $elements);

    return $form;
  }

  /**
   * Validate format of input fields and make sure the Sync Core backend is
   * accessible to actually update it.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    parent::validateForm($form, $form_state);

    $step = $this->getLastFormStep($form_state);

    $url = $form_state->getValue('backend_url');

    $client = SyncCore::getClient($url);
    $allowed_length = 0;
    try {
      $body = $client->getStatus();

      // In MongoDB (used by the Sync Core, the name of the database + collection name mustn't be longer than 120 characters)
      // TODO: Also validate before exporting (in case it's configured via config and exported via drush)
      $longest_entity_name = 0;
      $bundleInfoService = \Drupal::service('entity_type.bundle.info');
      $entity_types = $bundleInfoService->getAllBundleInfo();
      foreach ($entity_types as $type_key => $entity_type) {
        foreach ($entity_type as $entity_bundle_name => $entity_bundle) {
          $length = strlen($type_key) + strlen($entity_bundle_name) + 1;
          if ($length > $longest_entity_name) {
            $longest_entity_name = $length;
          }
        }
      }

      $allowed_length = isset($body['max_collection_name_length']) ? $body['max_collection_name_length'] : 120 - 7;

      // Full name is formatted like this: [database].drupal-[pool]-[site]-[entity]-[bundle].
      $allowed_length = $allowed_length - 6 - 1 - 1 - 1 - $longest_entity_name - 1;
    }
    catch (SyncCoreException $e) {
      $form_state->setErrorByName('backend_url', $this->t('The Sync Core is not accessible (did not respond with 200 OK). Please configure your outbound traffic to allow traffic to the Sync Core. The error message is @message', ['@message' => $e->getMessage()]));
    }

    if ($step === self::STEP_SYNC_CORE) {
      $methods = [
        Client::METHOD_POST,
        Client::METHOD_PATCH,
        Client::METHOD_DELETE,
      ];

      $auth_type = $form_state->getValue('authentication_type');

      $export_url = ContentSyncSettings::getInstance()->getSiteBaseUrl();

      $authentication_provider = AuthenticationByUser::getInstance();

      if ($auth_type == Pool::AUTHENTICATION_TYPE_BASIC_AUTH) {
        $authentication = [
          'type' => 'basic_auth',
          'username' => $authentication_provider->getUsername(),
          'password' => $authentication_provider->getPassword(),
          'base_url' => $export_url,
        ];
      }
      else {
        $authentication = [
          'type' => 'drupal8_services',
          'username' => $authentication_provider->getUsername(),
          'password' => $authentication_provider->getPassword(),
          'base_url' => $export_url,
        ];
      }

      foreach ($methods as $method) {
        try {
          $client->requestPing($export_url, $method, $authentication);
        }
        catch (ForbiddenException $e) {
          $form_state->setErrorByName('backend_url', $this->t('The Sync Core could not authenticate against this site. Please check that the Content Sync REST interface allows the configured authentication type and that the Content Sync user still exists. Message: @message', ['@message' => $e->getMessage()]));
          break;
        }
        catch (NotFoundException $e) {
          $form_state->setErrorByName('backend_url', $this->t('The Sync Core did not receive a valid response from this site for the method %method. Please configure inbound traffic to be allowed from the <a href="https://edge-box.atlassian.net/wiki/spaces/SUP/pages/143982665/Technical+Requirements">Sync Core IP addresses</a> to this site. Message: @message', ['@message' => $e->getMessage(), '%method' => $method]));
          break;
        }
        catch (BadRequestException $e) {
          $form_state->setErrorByName('backend_url', $this->t('The Sync Core could not reach this site because the URL of this site is invalid. Please update the site base URL in the Content Sync Settings tab. Raw IP addresses and localhost domain names are not allowed. Message: @message', ['@message' => $e->getMessage()]));
          break;
        }
        catch (SyncCoreException $e) {
          $form_state->setErrorByName('backend_url', $this->t('The Sync Core could not reach this site for the method %method. Please configure inbound traffic to be allowed from the <a href="https://edge-box.atlassian.net/wiki/spaces/SUP/pages/143982665/Technical+Requirements">Sync Core IP addresses</a> to this site. Message: @message', ['@message' => $e->getMessage(), '%method' => $method]));
          break;
        }
      }
    }

    if ($step === self::STEP_SITE_ID) {
      $site_id = $form_state->getValue('site_id');

      if (!preg_match('@^([a-z0-9\-_\.]+)$@', $site_id)) {
        $form_state->setErrorByName('site_id', $this->t('Please only use letters, numbers, underscores, dots and dashes.'));
      }

      if ($site_id == InstanceStorage::POOL_SITE_ID) {
        $form_state->setErrorByName('site_id', $this->t('This name is reserved.'));
      }
    }

    if ($step === self::STEP_POOL) {
      $api = $form_state->getValue('id');

      if (!preg_match('@^([a-z0-9\-_]+)$@', $api)) {
        $form_state->setErrorByName('id', $this->t('Please only use letters, numbers and dashes.'));
      }

      if ($api == 'drupal' || $api == 'api-unify') {
        $form_state->setErrorByName('api', $this->t('This name is reserved.'));
      }

      $site_id = $form_state->getValue('site_id');
      $length = strlen($api) + strlen($site_id);

      if ($length > $allowed_length) {
        $form_state->setErrorByName('id', $this->t('The Pool machine name + site id combined mustn\'t be longer than @allowed characters.', ['@allowed' => $allowed_length]));
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function actions(array $form, FormStateInterface $form_state) {
    $step = $this->getCurrentFormStep($form_state);

    if ($step !== self::STEP_POOL) {
      return [];
    }

    $element = parent::actions($form, $form_state);

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    /**
     * @var \Drupal\cms_content_sync\Entity\Pool $pool
     */
    $pool = $this->entity;

    if (!$pool->label()) {
      $pool->label = $this->getRemotePools($form_state)[$pool->id()];
    }

    $status = $pool->save();

    if ($status) {
      if (empty(Flow::getAll())) {
        $link = Link::createFromRoute(
          $this->t("Create a Flow"),
          "entity.cms_content_sync_flow.add_form"
        )->toString();
        \Drupal::messenger()->addStatus(
            $this->t('Saved Pool %label. Well done! @create to continue the setup.', [
              '%label' => $pool->label(),
              '@create' => $link,
            ])
        );
        \Drupal::messenger()->addStatus(
          $this->t('If you have connected another site already, @copy. Mirroring means you can simply swap the push and pull settings.', [
            '@copy' => Link::createFromRoute("copy or mirror the configuration from another site", "entity.cms_content_sync_flow.copy_remote")->toString(),
          ])
        );
      }
      else {

        \Drupal::messenger()->addStatus(
          $this->t('Saved Pool %label.', [
            '%label' => $pool->label(),
          ])
        );
      }
    }
    else {
      \Drupal::messenger()->addStatus(
        $this->t('The %label Pool could not be saved.', [
          '%label' => $pool->label(),
        ])
      );
    }

    // Make sure that the export is executed.
    $destination = \Drupal::request()->query->get('destination');
    \Drupal::request()->query->remove('destination');

    // Keep destination for after the export.
    $options = [];
    if (!empty($destination)) {
      $options['query']['destination'] = $destination;
    }

    $form_state->setRedirect('entity.cms_content_sync_pool.export', ['cms_content_sync_pool' => $this->entity->id()], $options);

  }

  /**
   * Helper function to check whether an Pool configuration entity exists.
   */
  public function exist($id) {
    $entity = $this->entityTypeManager->getStorage('cms_content_sync_pool')->getQuery()
      ->condition('id', $id)
      ->execute();
    return (bool) $entity;
  }

}
