<?php

namespace Drupal\cms_content_sync\Plugin\cms_content_sync\entity_handler;

use Drupal\cms_content_sync\Entity\Flow;
use Drupal\cms_content_sync\Exception\SyncException;
use Drupal\cms_content_sync\PushIntent;
use Drupal\cms_content_sync\PullIntent;
use Drupal\cms_content_sync\Plugin\EntityHandlerBase;
use Drupal\cms_content_sync\SyncIntent;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\crop\Entity\Crop;

/**
 * Class DefaultFileHandler, providing proper file handling capabilities.
 *
 * @EntityHandler(
 *   id = "cms_content_sync_default_file_handler",
 *   label = @Translation("Default File"),
 *   weight = 90
 * )
 *
 * @package Drupal\cms_content_sync\Plugin\cms_content_sync\entity_handler
 */
class DefaultFileHandler extends EntityHandlerBase {

  const USER_PROPERTY = 'uid';

  /**
   * @inheritdoc
   */
  public static function supports($entity_type, $bundle) {
    return $entity_type == 'file';
  }

  /**
   * @inheritdoc
   */
  public function getAllowedPushOptions() {
    return [
      PushIntent::PUSH_DISABLED,
      PushIntent::PUSH_AUTOMATICALLY,
      PushIntent::PUSH_AS_DEPENDENCY,
      PushIntent::PUSH_MANUALLY,
    ];
  }

  /**
   * @inheritdoc
   */
  public function getAllowedPreviewOptions() {
    return [
      'table' => 'Table',
      'preview_mode' => 'Preview mode',
    ];
  }

  /**
   * @inheritdoc
   */
  public function getHandlerSettings($current_values, $type = 'both') {
    $moduleHandler = \Drupal::service('module_handler');
    if ($moduleHandler->moduleExists('crop') && $moduleHandler->moduleExists('focal_point') && $type !== 'pull') {
      return [
        'export_crop' => [
          '#type' => 'checkbox',
          '#title' => 'Push cropping',
          '#default_value' => isset($current_values['export_crop']) && $current_values['export_crop'] === 0 ? 0 : 1,
        ],
      ];
    }
    else {
      return [];
    }
  }

  /**
   * @inheritdoc
   */
  public function updateEntityTypeDefinition(&$definition) {
    parent::updateEntityTypeDefinition($definition);

    $definition['new_properties']['apiu_file_content'] = [
      'type' => 'string',
      'default_value' => NULL,
    ];
    $definition['new_property_lists']['details']['apiu_file_content'] = 'value';
    $definition['new_property_lists']['filesystem']['apiu_file_content'] = 'value';
    $definition['new_property_lists']['modifiable']['apiu_file_content'] = 'value';
    $definition['new_property_lists']['required']['apiu_file_content'] = 'value';

    $definition['new_properties']['uri'] = [
      'type' => 'object',
      'default_value' => NULL,
      'multiple' => TRUE,
    ];
    $definition['new_property_lists']['details']['uri'] = 'value';
    $definition['new_property_lists']['database']['uri'] = 'value';
    $definition['new_property_lists']['required']['uri'] = 'value';
  }

  /**
   * @inheritdoc
   */
  public function getForbiddenFields() {
    return array_merge(
      parent::getForbiddenFields(),
      [
        'uri',
        'filemime',
        'filesize',
      ]
    );
  }

  /**
   * @inheritdoc
   */
  public function pull(PullIntent $intent) {
    /**
     * @var \Drupal\file\FileInterface $entity
     */
    $entity = $intent->getEntity();
    $action = $intent->getAction();

    if ($action == SyncIntent::ACTION_DELETE) {
      if ($entity) {
        return $this->deleteEntity($entity);
      }
      return FALSE;
    }

    $uri = $intent->getField('uri');
    if (empty($uri)) {
      throw new SyncException(SyncException::CODE_INVALID_PULL_REQUEST);
    }
    if (!empty($uri[0]['value'])) {
      $uri = $uri[0]['value'];
    }

    if (substr($uri, 0, 7) == 'http://' || substr($uri, 0, 8) == 'https://') {
      if (!$entity) {
        $entity_type = \Drupal::entityTypeManager()->getDefinition($intent->getEntityType());

        $base_data = [];

        if ($this->hasLabelProperty()) {
          $base_data[$entity_type->getKey('label')] = $intent->getField('title');
        }

        $base_data[$entity_type->getKey('uuid')] = $intent->getUuid();
        if ($entity_type->getKey('langcode')) {
          $base_data[$entity_type->getKey('langcode')] = $intent->getField($entity_type->getKey('langcode'));
        }

        $base_data['uri'] = $uri;

        $storage = \Drupal::entityTypeManager()->getStorage($intent->getEntityType());
        $entity = $storage->create($base_data);
      }

      $entity->set('filename', $intent->getField('title'));
      $entity->set('uri', $uri);
      $entity->save();

      return TRUE;
    }

    $content = $intent->getField('apiu_file_content');
    if (!$content) {
      throw new SyncException(SyncException::CODE_INVALID_PULL_REQUEST);
    }

    if ($entity) {
      // Drupal will re-use the existing file entity and keep it's ID, but
      // *change the UUID* of the file entity to a new random value
      // So we have to tell Drupal we actually want to keep it so references
      // to it keep working for us. That's why we can't use file_save_data- it doesn't do what it promises (keeping the
      // file entity and just replacing the file content).
      if ($uri = \Drupal::service('file_system')->saveData(base64_decode($content), $entity->getFileUri(), FileSystemInterface::EXISTS_REPLACE)) {
        $entity->set('filename', $intent->getField('title'));
        $entity->save();
        return TRUE;
      }

      throw new SyncException(SyncException::CODE_ENTITY_API_FAILURE);
    }
    else {
      $directory = \Drupal::service('file_system')->dirname($uri);
      $was_prepared = \Drupal::service('file_system')->prepareDirectory($directory, FileSystemInterface::CREATE_DIRECTORY);

      if ($was_prepared) {
        /* @var \Drupal\file\FileInterface[] $existing_files */
        $existing_files = \Drupal::entityTypeManager()
          ->getStorage('file')
          ->loadByProperties(['uri' => $uri]);

        $entity = file_save_data(base64_decode($content), $uri, FileSystemInterface::EXISTS_REPLACE);

        // Drupal has a pending issue: https://www.drupal.org/node/2241865
        // so it creates a new file entity even when overwriting an existing entity. This will throw an exception if we
        // now try to save the new file with the same UUID as the old file.
        // So if we're updating an existing file, we don't create a new file entity so just skipping the file safe.
        if (count($existing_files)) {
          $existing = reset($existing_files);
          // Yes, file exists and UUID matches. So no need to create a new file entity.
          if ($existing->uuid() === $intent->getUuid()) {
            // Delete duplicated file until Drupal resolves the issue above.
            if ($entity->uuid() !== $intent->getUuid()) {
              $entity->delete();
            }

            return TRUE;
          }
        }

        $entity->setPermanent();
        $entity->set('uuid', $intent->getUuid());
        $entity->set('filename', $intent->getField('title'));
        $entity->save();
        return TRUE;
      }

      throw new SyncException(SyncException::CODE_ENTITY_API_FAILURE);
    }
  }

  /**
   * @inheritdoc
   */
  public function push(PushIntent $intent, EntityInterface $entity = NULL) {
    /**
     * @var \Drupal\file\FileInterface $entity
     */
    if (!$entity) {
      $entity = $intent->getEntity();
    }

    if (!parent::push($intent)) {
      return FALSE;
    }

    // Base Info.
    $uri = $entity->getFileUri();
    $content = file_get_contents($uri);
    // File was removed from the file system. Trying to import it at another site will throw an error there and as the
    // source of the error is here, we throw an Error here.
    if ($content === FALSE) {
      \Drupal::logger('cms_content_sync')->error(
        'Can\'t push file: File @uri doesn\'t exist in the file system or the file permissions forbid access.',
        [
          '@uri' => $uri,
        ]
      );
      throw new \Exception("Can't push file: File " . $uri . " doesn't exist in the file system or the file permissions forbid access.");
    }

    $intent->setField('apiu_file_content', base64_encode($content));
    $intent->setField('uri', [['value' => $uri]]);
    $intent->setField('title', $entity->getFilename());

    // Preview.
    $view_mode = $this->flow->getPreviewType($entity->getEntityTypeId(), $entity->bundle());
    if ($view_mode != Flow::PREVIEW_DISABLED) {
      $intent->setField('preview', '<img style="max-height: 200px" src="' . file_create_url($uri) . '"/>');
    }
    else {
      $intent->setField('preview', '<em>Previews are disabled for this entity.</em>');
    }

    // No Translations, No Menu items compared to EntityHandlerBase.
    // Source URL.
    $this->setSourceUrl($intent, $entity);

    // Handle focal point crop entities.
    $moduleHandler = \Drupal::service('module_handler');
    if ($moduleHandler->moduleExists('crop') && $moduleHandler->moduleExists('focal_point')) {
      if ($this->settings['handler_settings']['export_crop']) {
        if (Crop::cropExists($uri, 'focal_point')) {
          $crop = Crop::findCrop($uri, 'focal_point');
          if ($crop) {
            $intent->embedEntity($crop, SyncIntent::ENTITY_REFERENCE_PUSH_AS_DEPENDENCY);

            $intent->setStatusData('crop', $crop->position());
          }
        }
      }
    }

    return TRUE;
  }

}
