<?php

namespace Drupal\cms_content_sync\Plugin\rest\resource;

use Drupal\cms_content_sync\Plugin\Type\EntityHandlerPluginManager;
use Drupal\cms_content_sync\SyncCore\Exception\SyncCoreException;
use Drupal\cms_content_sync\SyncCore\Query\Condition\DataCondition;
use Drupal\cms_content_sync\SyncCore\Query\Condition\ParentCondition;
use Drupal\cms_content_sync\SyncCore\Query\ListQuery;
use Drupal\cms_content_sync\SyncCore\Storage\EntityTypeStorage;
use Drupal\cms_content_sync\SyncCore\Storage\PreviewEntityStorage;
use Drupal\cms_content_sync\SyncIntent;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\EntityTypeBundleInfo;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\cms_content_sync\Entity\Flow;
use Drupal\cms_content_sync\Entity\EntityStatus;
use Drupal\cms_content_sync\Entity\Pool;
use Drupal\cms_content_sync\PullIntent;
use Drupal\rest\Plugin\ResourceBase;
use Drupal\Core\Render\Renderer;
use Drupal\rest\ResourceResponse;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides entity interfaces for Content Sync, allowing the manual
 * pull dashboard to query preview entities and to pull them to the site.
 *
 * @RestResource(
 *   id = "cms_content_sync_import_entity",
 *   label = @Translation("Content Sync Pull"),
 *   uri_paths = {
 *     "canonical" = "/rest/cms-content-sync-import/{pool}",
 *     "https://www.drupal.org/link-relations/create" = "/rest/cms-content-sync-import/{pool}/{entity_type_name}/{bundle_name}/{shared_entity_id}"
 *   }
 * )
 */
class PullEntity extends ResourceBase {

  /**
   * @var int CODE_INVALID_DATA The provided data could not be interpreted.
   */
  const CODE_INVALID_DATA = 401;

  /**
   * @var int CODE_NOT_FOUND The entity doesn't exist or can't be accessed
   */
  const CODE_NOT_FOUND = 404;

  /**
   * @var int CODE_UNEXPECTED_ERROR The Sync Core wasn't able to synchronize this entity.
   */
  const CODE_UNEXPECTED_ERROR = 500;

  /**
   * @var \Drupal\Core\Entity\EntityTypeBundleInfo
   */
  protected $entityTypeBundleInfo;

  /**
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  protected $entityTypeManager;

  /**
   * @var \Drupal\Core\Render\Renderer
   */
  protected $renderedManager;

  /**
   * @var \Drupal\Core\Entity\EntityRepositoryInterface
   */
  protected $entityRepository;

  /**
   * Constructs an object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param array $serializer_formats
   *   The available serialization formats.
   * @param \Psr\Log\LoggerInterface $logger
   *   A logger instance.
   * @param \Drupal\Core\Entity\EntityTypeBundleInfo $entity_type_bundle_info
   *   An entity type bundle info instance.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   An entity type manager instance.
   * @param \Drupal\Core\Render\Renderer $render_manager
   *   A rendered instance.
   * @param \Drupal\Core\Entity\EntityRepositoryInterface $entity_repository
   *   The entity repository interface.
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    array $serializer_formats,
    LoggerInterface $logger,
    EntityTypeBundleInfo $entity_type_bundle_info,
    EntityTypeManagerInterface $entity_type_manager,
    Renderer $render_manager,
    EntityRepositoryInterface $entity_repository
  ) {
    parent::__construct(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $serializer_formats,
      $logger
    );

    $this->entityTypeBundleInfo = $entity_type_bundle_info;
    $this->entityTypeManager = $entity_type_manager;
    $this->renderedManager = $render_manager;
    $this->entityRepository = $entity_repository;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition
  ) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->getParameter('serializer.formats'),
      $container->get('logger.factory')->get('rest'),
      $container->get('entity_type.bundle.info'),
      $container->get('entity_type.manager'),
      $container->get('renderer'),
      $container->get('entity.repository')
    );
  }

  /**
   * Responds to entity GET requests.
   *
   * @param string $pool_id
   *   The ID of the selected flow.
   *
   * @return \Symfony\Component\HttpFoundation\Response
   *   A list of entities of the given type and bundle.
   *
   * @throws \Exception
   */
  public function get($pool_id) {
    $cache_build = [
      '#cache' => [
        'max-age' => 0,
      ],
    ];

    // Loaded directly from the Sync Core, so we just need to provide the status for them.
    if (!empty($_GET['entities'])) {
      $items = [];

      $types = explode(';', $_GET['entities']);
      foreach ($types as $type) {
        list($type, $ids) = explode(':', $type);
        $ids = explode(',', $ids);
        foreach ($ids as $id) {
          $item = [
            'entity_type_id' => $type,
            'id' => $id,
          ];
          $this->enrichPreviewItem($item);
          $items[] = $item;
        }
      }

      $resource_response = new ResourceResponse([
        'items' => $items,
      ]);
      $resource_response->addCacheableDependency($cache_build);

      return $resource_response;
    }

    $pool = Pool::getAll()[$pool_id];

    if (!$pool) {
      $resource_response = new ResourceResponse(['message' => "Unknown pool $pool_id."], self::CODE_NOT_FOUND);
      $resource_response->addCacheableDependency($cache_build);
      return $resource_response;
    }

    $entity_type_ids = [];
    $entity_type_name = isset($_GET['entity_type_name']) ? $_GET['entity_type_name'] : NULL;
    $bundle_name = isset($_GET['bundle_name']) ? $_GET['bundle_name'] : NULL;
    $page = isset($_GET['page']) ? intval($_GET['page']) : 0;

    $filters = [];

    /**
     * @var \Drupal\Core\Entity\EntityFieldManager $entityFieldManager
     */
    $entityFieldManager = \Drupal::service('entity_field.manager');

    foreach (Flow::getAll() as $flow) {
      foreach ($flow->getEntityTypeConfig() as $definition) {
        if (!$flow->canPullEntity($definition['entity_type_name'], $definition['bundle_name'], PullIntent::PULL_MANUALLY)) {
          continue;
        }

        if ($entity_type_name && $definition['entity_type_name'] != $entity_type_name) {
          continue;
        }

        if ($bundle_name && $definition['bundle_name'] != $bundle_name) {
          continue;
        }

        if (empty($definition['import_pools'][$pool->id]) || $definition['import_pools'][$pool->id] != Pool::POOL_USAGE_ALLOW) {
          continue;
        }

        $id = EntityTypeStorage::getExternalEntityTypeId(
          $pool->id,
          $definition['entity_type_name'],
          $definition['bundle_name']
        );
        if (in_array($id, $entity_type_ids)) {
          continue;
        }

        /**
         * @var \Drupal\Core\Field\FieldDefinitionInterface[] $fields
         */
        $fields = $entityFieldManager->getFieldDefinitions($definition['entity_type_name'], $definition['bundle_name']);
        foreach ($fields as $key => $field) {
          $field_config = $flow->getFieldHandlerConfig($definition['entity_type_name'], $definition['bundle_name'], $key);
          if (empty($field_config)) {
            continue;
          }

          if (empty($field_config['handler_settings']['subscribe_only_to'])) {
            continue;
          }

          $allowed = [];

          foreach ($field_config['handler_settings']['subscribe_only_to'] as $ref) {
            $allowed[] = $ref['uuid'];
          }

          $filters[] = ParentCondition::any()
            ->add(
              ParentCondition::none()
                ->add(DataCondition::equal(PreviewEntityStorage::PROPERTY_ENTITY_TYPE_UNVERSIONED, $id))
            )
            ->add(DataCondition::in(PreviewEntityStorage::PROPERTY_CUSTOM_PROPERTIES . '.' . $key . '.' . SyncIntent::UUID_KEY, $allowed));
        }

        $entity_type_ids[] = $id;
      }
    }

    if (empty($entity_type_ids)) {
      $resource_response = new ResourceResponse(['message' => 'No previews available.'], self::CODE_NOT_FOUND);
      $resource_response->addCacheableDependency($cache_build);
      return $resource_response;
    }

    $order_field = isset($_GET['order_by']) && $_GET['order_by'] == 'title' ? 'title' : 'published_date';

    $condition = DataCondition::in(PreviewEntityStorage::PROPERTY_ENTITY_TYPE_UNVERSIONED, $entity_type_ids);

    if (!empty($_GET['filter_title'])) {
      $filters[] = DataCondition::search(PreviewEntityStorage::PROPERTY_TITLE, $_GET['filter_title']);
    }
    if (!empty($_GET['filter_preview'])) {
      $filters[] = DataCondition::search(PreviewEntityStorage::PROPERTY_PREVIEW_HTML, $_GET['filter_preview']);
    }
    if (!empty($_GET['filter_published_from'])) {
      $filters[] = DataCondition::greaterOrEqual(PreviewEntityStorage::PROPERTY_PUBLISHED_DATE, (int) $_GET['filter_published_from']);
    }
    if (!empty($_GET['filter_published_to'])) {
      // Include last allowed date.
      $filters[] = DataCondition::lessOrEqual(PreviewEntityStorage::PROPERTY_PUBLISHED_DATE, (int) $_GET['filter_published_to'] + 24 * 60 * 60);
    }

    if (count($filters)) {
      $filters[] = $condition;
      $condition = ParentCondition::all($filters);
    }

    try {
      $data = $pool
        ->getClient()
        ->getPreviewEntityStorage()
        ->listItems()
        ->setItemsPerPage(10)
        ->getDetails()
        ->setCondition($condition)
        ->orderBy($order_field,
          isset($_GET['order_direction']) && $_GET['order_direction'] == 'asc' ?
            ListQuery::ORDER_ASCENDING :
            ListQuery::ORDER_DESCENDING
        )
        ->setPage($page ? $page : 1)
        ->execute()
        ->getRaw();
    }
    catch (SyncCoreException $e) {
      $body = $e->getResponseBody();
      $code = $e->getStatusCode();
      $resource_response = new ResourceResponse($body ? $body : json_encode(['message' => $e->getMessage()]), $code ? $code : 500);
      $resource_response->addCacheableDependency($cache_build);
      return $resource_response;
    }

    foreach ($data['items'] as &$item) {
      $this->enrichPreviewItem($item);
    }

    $resource_response = new ResourceResponse($data);
    $resource_response->addCacheableDependency($cache_build);

    return $resource_response;
  }

  /**
   * Add entity status for the entity in question. Add "entity_type_name"
   * and "bundle" information.
   *
   * @param $item
   *
   * @throws \Exception
   */
  protected function enrichPreviewItem(&$item) {
    $shared_entity_id = $item['id'];

    $entity_type_id = $item['entity_type_id'];
    list(,, $entity_type_name, $bundle_name) = explode('-', $entity_type_id);

    $item['entity_type_name'] = $entity_type_name;
    $item['bundle_name']      = $bundle_name;
    $item['entity_status']    = [];
    $item['last_import']      = NULL;
    $item['last_export']      = NULL;
    $item['deleted']          = FALSE;
    $item['is_source']        = FALSE;

    /**
     * @var \Drupal\Core\Entity\EntityInterface $entity
     */
    if (EntityHandlerPluginManager::isEntityTypeConfiguration($entity_type_name)) {
      $entity = \Drupal::entityTypeManager()
        ->getStorage($entity_type_name)
        ->load($shared_entity_id);
      $entity_uuid = NULL;
    }
    else {
      $entity = \Drupal::service('entity.repository')
        ->loadEntityByUuid($entity_type_name, $shared_entity_id);
      $entity_uuid = $shared_entity_id;
    }

    if ($entity) {
      if ($entity->hasLinkTemplate('canonical')) {
        try {
          $url = $entity->toUrl('canonical', ['absolute' => TRUE])
            ->toString(TRUE)
            ->getGeneratedUrl();
          $item['local_url'] = $url;
        }
        catch (\Exception $e) {
        }
      }

      $entity_uuid = $entity->uuid();
    }

    if (!$entity_uuid) {
      return;
    }

    $entity_status = EntityStatus::getInfosForEntity($entity_type_name, $entity_uuid);

    foreach ($entity_status as $info) {
      $item['entity_status'][] = [
        'flow_id' => $info->get('flow')->value,
        'pool_id' => $info->get('pool')->value,
        'last_import' => $info->getLastPull(),
        'last_export' => $info->getLastPush(),
      ];
      if (!$item['last_import'] || $item['last_import'] < $info->getLastPull()) {
        $item['last_import'] = $info->getLastPull();
      }
      if (!$item['last_export'] || $item['last_export'] < $info->getLastPush()) {
        $item['last_export'] = $info->getLastPush();
      }
      if ($info->isDeleted()) {
        $item['deleted'] = TRUE;
      }
      if ($info->isSourceEntity()) {
        $item['is_source'] = TRUE;
      }
    }

    // We had a bug before where the "deleted" flag wasn't set correctly locally. So we're mitigating for that here.
    if ($item['last_import'] && !$entity) {
      $item['deleted'] = TRUE;
    }
  }

  /**
   * Responds to entity POST requests.
   *
   * @param string $pool_id
   * @param string $entity_type_name
   * @param string $bundle_name
   * @param string $shared_entity_id
   *
   * @return \Symfony\Component\HttpFoundation\Response
   *   -
   *
   * @throws \Exception
   */
  public function post($pool_id, $entity_type_name, $bundle_name, $shared_entity_id) {
    $pool = Pool::getAll()[$pool_id];

    if (!$pool) {
      return new ResourceResponse(['message' => "Unknown pool ID $pool_id."], self::CODE_NOT_FOUND);
    }

    $success = FALSE;

    foreach (Flow::getAll() as $flow) {
      foreach ($flow->getEntityTypeConfig() as $definition) {
        if (!$flow->canPullEntity($definition['entity_type_name'], $definition['bundle_name'], PullIntent::PULL_MANUALLY)) {
          continue;
        }
        if ($definition['entity_type_name'] != $entity_type_name) {
          continue;
        }
        if ($definition['bundle_name'] != $bundle_name) {
          continue;
        }
        if ($definition['import_pools'][$pool->id] != Pool::POOL_USAGE_ALLOW) {
          continue;
        }

        /**
         * @var \Drupal\cms_content_sync\SyncCore\Entity\ConnectionSynchronization $sync
         */
        $sync = $pool->getConnectionSynchronizationForEntityType(
          $entity_type_name,
          $bundle_name,
          $definition['version'],
          FALSE
        );

        $success = $sync
          ->synchronizeSingle($shared_entity_id)
          ->isManual(TRUE)
          ->execute()
          ->succeeded();

        if (!$success) {
          return new ResourceResponse(['message' => 'Failed to pull entity.'], self::CODE_UNEXPECTED_ERROR);
        }
      }
    }

    if (!$success) {
      return new ResourceResponse(['message' => "Missing flow for pool $pool_id."], self::CODE_NOT_FOUND);
    }

    /**
     * @var \Drupal\cms_content_sync\SyncCore\Storage\PreviewEntityStorage $previews
     */
    $previews = $pool
      ->getClient()
      ->getPreviewEntityStorage();

    $query = $previews
      ->getItem($shared_entity_id)
      ->execute();

    if (!$query->succeeded()) {
      return new ResourceResponse(['message' => 'Failed to load preview.'], self::CODE_UNEXPECTED_ERROR);
    }

    $data = $query->getItem();

    $this->enrichPreviewItem($data);

    // Entity was ignored by the Sync Core for some reason or the update was rejected by Drupal.
    if (empty($data['last_import'])) {
      return new ResourceResponse(['message' => 'Failed to pull entity.'], self::CODE_UNEXPECTED_ERROR);
    }

    return new ResourceResponse($data);
  }

}
