<?php

namespace Drupal\cms_content_sync;

use Drupal\cms_content_sync\Event\AfterEntityPush;
use Drupal\cms_content_sync\SyncCore\Exception\NotFoundException;
use Drupal\cms_content_sync\SyncCore\Exception\SyncCoreException;
use Drupal\cms_content_sync\SyncCore\Storage\CustomStorage;
use Drupal\cms_content_sync\SyncCore\Storage\InstanceStorage;
use Drupal\Core\Config\Entity\ConfigEntityInterface;
use Drupal\Core\Entity\EntityChangedInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\TranslatableInterface;
use Drupal\cms_content_sync\Entity\Flow;
use Drupal\cms_content_sync\Entity\EntityStatus;
use Drupal\cms_content_sync\Entity\Pool;
use Drupal\cms_content_sync\Exception\SyncException;
use Drupal\crop\Entity\Crop;

/**
 * Class PushIntent.
 *
 * @package Drupal\cms_content_sync
 */
class PushIntent extends SyncIntent {
  /**
   * @var string PUSH_DISABLED
   *   Disable pushing completely for this entity type, unless forced.
   *   - used as a configuration option
   *   - not used as $action
   */
  const PUSH_DISABLED = 'disabled';
  /**
   * @var string PUSH_AUTOMATICALLY
   *   Automatically push all entities of this entity type.
   *   - used as a configuration option
   *   - used as $action
   */
  const PUSH_AUTOMATICALLY = 'automatically';
  /**
   * @var string PUSH_MANUALLY
   *   Push only some of these entities, chosen manually.
   *   - used as a configuration option
   *   - used as $action
   */
  const PUSH_MANUALLY = 'manually';
  /**
   * @var string PUSH_AS_DEPENDENCY
   *   Push only some of these entities, pushed if other pushed entities
   *   use it.
   *   - used as a configuration option
   *   - used as $action
   */
  const PUSH_AS_DEPENDENCY = 'dependency';
  /**
   * @var string PUSH_FORCED
   *   Force the entity to be pushed (as long as a handler is also selected).
   *   Can be used programmatically for custom workflows.
   *   - not used as a configuration option
   *   - used as $action
   */
  const PUSH_FORCED = 'forced';
  /**
   * @var string PUSH_ANY
   *   Only used as a filter to check if the Flow pushes this entity in any
   *   way.
   *   - not used as a configuration option
   *   - not used as $action
   *   - so only used to query against Flows that have *any* push setting for a given entity (type).
   */
  const PUSH_ANY = 'any';

  /**
   * @var string PUSH_FAILED_REQUEST_FAILED
   *   The request to the Sync Core failed completely.
   */
  const PUSH_FAILED_REQUEST_FAILED = 'export_failed_request_failed';
  /**
   * @var string PUSH_FAILED_REQUEST_INVALID_STATUS_CODE
   *   The Sync Core returned a non-2xx status code.
   */
  const PUSH_FAILED_REQUEST_INVALID_STATUS_CODE = 'export_failed_invalid_status_code';
  /**
   * @var string PUSH_FAILED_DEPENDENCY_PUSH_FAILED
   *   The entity wasn't pushed because when pushing a dependency, an error was thrown.
   */
  const PUSH_FAILED_DEPENDENCY_PUSH_FAILED = 'export_failed_dependency_export_failed';
  /**
   * @var string PUSH_FAILED_INTERNAL_ERROR
   *   The entity wasn't pushed because when serializing it, an error was thrown.
   */
  const PUSH_FAILED_INTERNAL_ERROR = 'export_failed_internal_error';
  /**
   * @var string PUSH_FAILED_HANDLER_DENIED
   *   Soft fail: The push failed because the handler returned FALSE when executing the push.
   */
  const PUSH_FAILED_HANDLER_DENIED = 'export_failed_handler_denied';
  /**
   * @var string PUSH_FAILED_UNCHANGED
   *   Soft fail: The entity wasn't pushed because it didn't change since the last push.
   */
  const PUSH_FAILED_UNCHANGED = 'export_failed_unchanged';

  /**
   * PushIntent constructor.
   *
   * @param \Drupal\cms_content_sync\Entity\Flow $flow
   * @param \Drupal\cms_content_sync\Entity\Pool $pool
   * @param $reason
   * @param $action
   * @param \Drupal\Core\Entity\EntityInterface $entity
   */
  public function __construct(Flow $flow, Pool $pool, $reason, $action, EntityInterface $entity) {
    parent::__construct($flow, $pool, $reason, $action, $entity->getEntityTypeId(), $entity->bundle(), $entity->uuid(), $entity instanceof ConfigEntityInterface ? $entity->id() : NULL);

    if (!$this->entity_status->getLastPush()) {
      if (!EntityStatus::getLastPullForEntity($entity) && !PullIntent::entityHasBeenPulledFromRemoteSite($entity->getEntityTypeId(), $entity->uuid())) {
        $this->entity_status->isSourceEntity(TRUE);
      }
    }

    $this->entity = $entity;

    $moduleHandler = \Drupal::service('module_handler');
    $quickedit_enabled = $moduleHandler->moduleExists('quickedit');
    if ($quickedit_enabled && !empty(\Drupal::service('tempstore.private')->get('quickedit')->get($entity->uuid()))) {
      $this->isQuickEdited = TRUE;
    }
  }

  /**
   * Get the correct synchronization for a specific action on a given entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   * @param string|string[] $reason
   * @param string $action
   *
   * @return \Drupal\cms_content_sync\Entity\Flow[]
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public static function getFlowsForEntity(EntityInterface $entity, $reason, $action = SyncIntent::ACTION_CREATE) {
    $flows = Flow::getAll();

    $result = [];

    foreach ($flows as $flow) {
      if ($flow->canPushEntity($entity, $reason, $action)) {
        $result[] = $flow;
      }
    }

    return $result;
  }

  /**
   * Get the correct synchronization for a specific action on a given entity.
   *
   * @param string $entity_type_name
   * @param string|null $bundle_name
   * @param string|string[] $reason
   * @param string $action
   *
   * @return \Drupal\cms_content_sync\Entity\Flow[]
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public static function getFlowsForEntityType($entity_type_name, $bundle_name, $reason, $action = SyncIntent::ACTION_CREATE) {
    $flows = Flow::getAll();

    $result = [];

    foreach ($flows as $flow) {
      if ($flow->canPushEntityType($entity_type_name, $bundle_name, $reason, $action)) {
        $result[] = $flow;
      }
    }

    return $result;
  }

  /**
   * Serialize the given entity using the entity push and field push
   * handlers.
   *
   * @param array &$result
   *   The data to be provided to Sync Core.
   *
   * @throws \Drupal\cms_content_sync\Exception\SyncException
   *
   * @return bool
   *   Whether or not the serialized entity could be created.
   */
  public function serialize(array &$result) {
    $config = $this->flow->getEntityTypeConfig($this->entityType, $this->bundle);
    $handler = $this->flow->getEntityTypeHandler($config);

    $status = $handler->push($this);

    if (!$status) {
      return FALSE;
    }

    $result = $this->getData();
    return TRUE;
  }

  /**
   * Wrapper for {@see Flow::getExternalConnectionPath}.
   *
   * @param string $shared_entity_id
   * @param bool $as_dependency
   * @param string $version
   *   The version used for the push. For DELETE requests this may be a previous version. All
   *   other requests must use the newest version. If an UPDATE is sent with a new version, it becomes a CREATE instead.
   *
   * @return string
   */
  public function getExternalUrl($shared_entity_id, $as_dependency = FALSE, $version = NULL) {
    $url = $this->pool->getBackendUrl() . '/' . CustomStorage::getCustomPath(
        $this->pool->id,
        $this->pool->getSiteId(),
        $this->entityType,
        $this->bundle
      );

    if ($shared_entity_id) {
      $url .= '/' . $shared_entity_id;
    }

    if ($as_dependency) {
      $url .= '?is_dependency=true';
    }

    return $url;
  }

  /**
   * Save that the pull for the given entity failed.
   *
   * @param string $failure_reason
   *   See PushIntent::PUSH_FAILURE_*.
   * @param null|string $message
   *   An optional message accompanying this error.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  protected function saveFailedPush($failure_reason, $message = NULL) {
    $soft_fails = [
      PushIntent::PUSH_FAILED_HANDLER_DENIED,
      PushIntent::PUSH_FAILED_UNCHANGED,
    ];

    $soft = in_array($failure_reason, $soft_fails);

    $this->entity_status->didPushFail(TRUE, $soft, [
      'error' => $failure_reason,
      'action' => $this->getAction(),
      'reason' => $this->getReason(),
      'message' => $message,
    ]);

    $this->entity_status->save();
  }

  /**
   * @return int
   */
  protected function getRequestTime() {
    return (int) $_SERVER['REQUEST_TIME'];
  }

  protected $isQuickEdited = FALSE;

  /**
   * Get the changed date of the entity. Not all entities provide the required attribute and even those aren't
   * consistently saving it so this method takes care of these exceptions.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *
   * @return int
   */
  protected function getEntityChangedTime($entity) {
    $request_time = $this->getRequestTime();
    $push = $request_time;

    if ($entity instanceof EntityChangedInterface) {
      $push = $entity->getChangedTime();
      if ($entity instanceof TranslatableInterface) {
        foreach ($entity->getTranslationLanguages(FALSE) as $language) {
          $translation = $entity->getTranslation($language->getId());
          /**
           * @var \Drupal\Core\Entity\EntityChangedInterface $translation
           */
          if ($translation->getChangedTime() > $push) {
            $push = $translation->getChangedTime();
          }
        }
      }
      // Check if any bricks were updated during this request that this specific entity is referencing
      // Quick edit doesn't update the changed date of the node...... so we have to go and see manually if anything
      // changed by caching it....
      if ($push < $request_time && $this->isQuickEdited) {
        return $request_time;
      }
    }

    /**
     * @var \Drupal\Core\Entity\EntityFieldManager $entity_field_manager
     */
    $entity_field_manager = \Drupal::service('entity_field.manager');

    /**
     * @var \Drupal\Core\Field\FieldDefinitionInterface[] $fields
     */
    $fields = $entity_field_manager->getFieldDefinitions($entity->getEntityTypeId(), $entity->bundle());

    // Elements that are inline edited in other forms like media elements edited within node forms don't get their
    // timestamp updated even though the file attributes may change like the focal point. So we are checking for file
    // reference fields and check if they were updated in the meantime.
    foreach ($fields as $key => $field) {
      if ($field->getType() === 'image') {
        $data = $entity->get($key)->getValue();

        foreach ($data as $delta => $value) {
          if (empty($value['target_id'])) {
            continue;
          }

          $entityTypeManager = \Drupal::entityTypeManager();
          $storage = $entityTypeManager->getStorage('file');

          $target_id = $value['target_id'];
          $reference = $storage->load($target_id);

          if (!$reference) {
            continue;
          }

          $sub = $this->getEntityChangedTime($reference);
          if ($sub > $push) {
            $push = $sub;
          }
        }
      }
    }

    // File entities timestamp doesn't change when focal point is updated and crop entity doesn't provide a changed date.
    if ($entity->getEntityTypeId() === 'file') {
      /**
       * @var \Drupal\file\FileInterface $entity
       */

      // Handle focal point crop entities.
      $moduleHandler = \Drupal::service('module_handler');
      if ($moduleHandler->moduleExists('crop') && $moduleHandler->moduleExists('focal_point')) {
        if (Crop::cropExists($entity->getFileUri(), 'focal_point')) {
          $crop = Crop::findCrop($entity->getFileUri(), 'focal_point');
          if ($crop) {
            $info = EntityStatus::getInfoForEntity('file', $entity->uuid(), $this->flow->id(), $this->pool->id());

            if ($info) {
              $position = $crop->position();
              $last = $info->getData('crop');
              if (empty($last) || $position['x'] !== $last['x'] || $position['y'] !== $last['y']) {
                $push = $this->getRequestTime();
              }
            }
          }
        }
      }
    }

    return $push;
  }

  /**
   * Push the given entity.
   *
   * @param bool $return_only
   *
   * @return array|bool TRUE|FALSE if the entity is pushed via REST.
   *   NULL|string (serialized entity) if $return_only is set to TRUE.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\cms_content_sync\Exception\SyncException
   * @throws \GuzzleHttp\Exception\GuzzleException
   */
  public function execute($return_only = FALSE) {
    $action = $this->getAction();
    $reason = $this->getReason();
    $entity = $this->getEntity();

    /**
     * @var array $deletedTranslations
     *   The translations that have been deleted. Important to notice when
     *   updates must be performed (see ::ACTION_DELETE_TRANSLATION).
     */
    static $deletedTranslations = [];

    if ($action == SyncIntent::ACTION_DELETE_TRANSLATION) {
      $deletedTranslations[$entity->getEntityTypeId()][$entity->uuid()] = TRUE;
      return FALSE;
    }

    if ($entity instanceof TranslatableInterface) {
      $entity = $entity->getUntranslated();
      $this->entity = $entity;
    }
    $push = $this->getEntityChangedTime($entity);

    // If this very request was sent to delete/create this entity, ignore the
    // push as the result of this request will already tell Sync Core it has
    // been deleted. Otherwise Sync Core will return a reasonable 404 for
    // deletions.
    if (PullIntent::entityHasBeenPulledFromRemoteSite($entity->getEntityTypeId(), $entity->uuid())) {
      self::$noPushReasons[$entity->getEntityTypeId()][$entity->uuid()] = self::NO_PUSH_REASON__JUST_PULLED;
      return FALSE;
    }

    $entity_type   = $entity->getEntityTypeId();
    $entity_bundle = $entity->bundle();
    $entity_uuid   = $entity->uuid();

    $pushed = $this->entity_status->getLastPush();

    if ($pushed) {
      if ($action == SyncIntent::ACTION_CREATE) {
        $action = SyncIntent::ACTION_UPDATE;
      }
    }
    else {
      if ($action == SyncIntent::ACTION_UPDATE) {
        $action = SyncIntent::ACTION_CREATE;
      }
      // If the entity was deleted but has never been pushed before,
      // pushing the deletion action doesn't make sense as it doesn't even
      // exist remotely.
      elseif ($action == SyncIntent::ACTION_DELETE) {
        self::$noPushReasons[$entity->getEntityTypeId()][$entity->uuid()] = self::NO_PUSH_REASON__NEVER_PUSHED;
        return FALSE;
      }
    }

    $cms_content_sync_disable_optimization = boolval(\Drupal::config('cms_content_sync.debug')
      ->get('cms_content_sync_disable_optimization'));

    if (!self::$pushed) {
      self::$pushed = [];
    }
    if (isset(self::$pushed[$action][$entity_type][$entity_bundle][$entity_uuid][$this->pool->id]) && !$return_only) {
      return self::$pushed[$action][$entity_type][$entity_bundle][$entity_uuid][$this->pool->id];
    }
    if ($action == SyncIntent::ACTION_CREATE) {
      if (isset(self::$pushed[SyncIntent::ACTION_UPDATE][$entity_type][$entity_bundle][$entity_uuid][$this->pool->id]) && !$return_only) {
        return self::$pushed[SyncIntent::ACTION_UPDATE][$entity_type][$entity_bundle][$entity_uuid][$this->pool->id];
      }
    }
    elseif ($action == SyncIntent::ACTION_UPDATE) {
      if (isset(self::$pushed[SyncIntent::ACTION_CREATE][$entity_type][$entity_bundle][$entity_uuid][$this->pool->id]) && !$return_only) {
        return self::$pushed[SyncIntent::ACTION_CREATE][$entity_type][$entity_bundle][$entity_uuid][$this->pool->id];
      }
    }
    self::$pushed[$action][$entity_type][$entity_bundle][$entity_uuid][$this->pool->id] = TRUE;

    // If the entity didn't change, it doesn't have to be pushed again.
    if (!$cms_content_sync_disable_optimization && $this->entity_status->getLastPush() && $this->entity_status->getLastPush() >= $push && $reason != self::PUSH_FORCED &&
      $action != SyncIntent::ACTION_DELETE &&
      empty($deletedTranslations[$entity->getEntityTypeId()][$entity->uuid()])) {

      // Don't use optimization for taxonomy terms as Drupal doesn't update the
      // changed timestamp on the entity when moving it in the tree for the
      // first time.
      // Menu link translations share their change time stamp with the original
      // entity, so translations would not be pushed.
      // Same issue when using draggable views- timestamp of the related entities isn't saved and draggableviews also
      // doesn't provide any timestamps.
      if ($entity_type != 'taxonomy_term' && $entity_type != 'menu_link_content') {
        self::$noPushReasons[$entity->getEntityTypeId()][$entity->uuid()] = self::NO_PUSH_REASON__UNCHANGED;
        return FALSE;
      }
    }

    /** @var \Drupal\Core\Entity\EntityRepositoryInterface $entity_repository */
    $entity_repository = \Drupal::service('entity.repository');

    $proceed = TRUE;
    $body = NULL;

    if ($action != SyncIntent::ACTION_DELETE) {
      $body = [];

      try {
        $proceed = $this->serialize($body);
      }
      catch (\Exception $e) {
        $this->saveFailedPush(PushIntent::PUSH_FAILED_INTERNAL_ERROR, $e->getMessage());

        throw new SyncException(SyncException::CODE_ENTITY_API_FAILURE, $e);
      }

      if ($proceed) {
        $embedded_entities = [];
        if (!empty($body['embed_entities'])) {
          foreach ($body['embed_entities'] as $data) {
            try {
              /**
               * @var \Drupal\Core\Entity\EntityInterface $embed_entity
               */
              $embed_entity = $entity_repository->loadEntityByUuid($data[SyncIntent::ENTITY_TYPE_KEY], $data[SyncIntent::UUID_KEY]);
              $all_pools    = Pool::getAll();
              $pools        = $this->flow->getPoolsToPushTo($entity, $this->getReason(), $this->getAction(), TRUE);
              $used_pools   = [];

              $flows = Flow::getAll();
              $flows = [$this->flow->id => $this->flow] + $flows;

              $version = Flow::getEntityTypeVersion($embed_entity->getEntityTypeId(), $embed_entity->bundle());

              $export_dependency = function ($reason)
 use ($flows, $embed_entity, &$used_pools, $data, $pools, $version, $all_pools, &$embedded_entities) {
                foreach ($flows as $flow) {
                  if (!$flow->canPushEntity($embed_entity, $reason, SyncIntent::ACTION_CREATE)) {
                    continue;
                  }

                  foreach ($flow->getEntityTypeConfig($embed_entity->getEntityTypeId(), $embed_entity->bundle())['export_pools'] as $pool_id => $behavior) {
                    if (in_array($pool_id, $used_pools)) {
                      continue;
                    }

                    if ($behavior == Pool::POOL_USAGE_FORBID) {
                      continue;
                    }

                    // If this entity was newly created, it won't have any groups to push to
                    // selected, unless they're FORCED. In this case we add default sync
                    // groups based on the parent entity, as you would expect.
                    if ($data[SyncIntent::AUTO_PUSH_KEY]) {
                      if (!isset($pools[$pool_id])) {
                        // TODO: Save all parent > child relationships so we can check if this pool is used somewhere else.
                        continue;
                      }

                      $pool = $pools[$pool_id];
                      $info = EntityStatus::getInfoForEntity($embed_entity->getEntityTypeId(), $embed_entity->uuid(), $flow, $pool);

                      if (!$info) {
                        $info = EntityStatus::create([
                          'flow' => $flow->id,
                          'pool' => $pool->id,
                          'entity_type' => $embed_entity->getEntityTypeId(),
                          'entity_uuid' => $embed_entity->uuid(),
                          'entity_type_version' => $version,
                          'flags' => 0,
                        ]);
                      }

                      $info->isPushEnabled(NULL, TRUE);
                      $info->save();
                    }
                    else {
                      $pool = $all_pools[$pool_id];
                      if ($behavior == Pool::POOL_USAGE_ALLOW) {
                        $info = EntityStatus::getInfoForEntity($embed_entity->getEntityTypeId(), $embed_entity->uuid(), $flow, $pool);
                        if (!$info || !$info->isPushEnabled()) {
                          continue;
                        }
                      }
                    }

                    PushIntent::pushEntity($embed_entity, $reason, SyncIntent::ACTION_CREATE, $flow, $pool);

                    $info = EntityStatus::getInfoForEntity($embed_entity->getEntityTypeId(), $embed_entity->uuid(), $flow, $pool);
                    if (!$info || !$info->getLastPush()) {
                      continue;
                    }

                    $used_pools[] = $pool_id;
                    $definition = $data;
                    $definition[SyncIntent::API_KEY] = $pool->id;
                    $definition[SyncIntent::SOURCE_CONNECTION_ID_KEY] = CustomStorage::getCustomId(
                      $pool->id,
                      $pool->getSiteId(),
                      $embed_entity->getEntityTypeId(),
                      $embed_entity->bundle()
                    );
                    $definition[SyncIntent::POOL_CONNECTION_ID_KEY] = CustomStorage::getCustomId(
                      $pool->id,
                      InstanceStorage::POOL_SITE_ID,
                      $embed_entity->getEntityTypeId(),
                      $embed_entity->bundle()
                    );
                    $embedded_entities[] = $definition;
                  }
                }
              };

              // Prefer Flows that push AS_DEPENDENCY.
              $export_dependency(self::PUSH_AS_DEPENDENCY);

              // But then use Flows that push AUTOMATICALLY to support that use-case as well.
              // If an AS_DEPENDENCY Flow has pushed to a specific pool before, the Flow pushing AUTOMATICALLY will not
              // be able to export to that pool again here, that's why the order matters.
              $export_dependency(self::PUSH_AUTOMATICALLY);
            }
            catch (\Exception $e) {
              $this->saveFailedPush(PushIntent::PUSH_FAILED_DEPENDENCY_PUSH_FAILED, $e->getMessage());

              throw new SyncException(SyncException::CODE_UNEXPECTED_EXCEPTION, $e);
            }
          }
        }

        $body['embed_entities'] = $embedded_entities;
      }
    }

    \Drupal::logger('cms_content_sync')->info('@not @embed PUSH @action @entity_type:@bundle @uuid @reason: @message', [
      '@reason' => $reason,
      '@action' => $action,
      '@entity_type'  => $entity_type,
      '@bundle' => $entity_bundle,
      '@uuid' => $entity_uuid,
      '@not' => $proceed ? '' : 'NO',
      '@embed' => $return_only ? 'EMBEDDING' : '',
      '@message' => $proceed ? t('The entity has been pushed.') : t('The entity handler denied to push this entity.'),
    ]);

    // Handler chose to deliberately ignore this entity,
    // e.g. a node that wasn't published yet and is not pushed unpublished.
    if (!$proceed) {
      self::$noPushReasons[$entity->getEntityTypeId()][$entity->uuid()] = self::NO_PUSH_REASON__HANDLER_IGNORES;
      $this->saveFailedPush(PushIntent::PUSH_FAILED_HANDLER_DENIED);
      return $return_only ? NULL : FALSE;
    }

    // We need to update the revision timestamp as otherwise the change won't be propagated by the Sync Core.
    if ($this->isQuickEdited) {
      if (!empty($body['revision_timestamp'][0]['value']) && $body['revision_timestamp'][0]['value'] < $this->getRequestTime()) {
        $body['revision_timestamp'][0]['value'] = $this->getRequestTime();
      }
    }

    if ($return_only) {
      self::$pushed[$action][$entity_type][$entity_bundle][$entity_uuid][$this->pool->id] = $body;
      return $body;
    }

    $version = $this->flow->getEntityTypeConfig($entity->getEntityTypeId(), $entity->bundle())['version'];

    // If the version changed, UPDATE becomes CREATE instead and DELETE requests must be performed against the old
    // version, as otherwise they would result in a 404 Not Found response.
    if ($version != $this->entity_status->getEntityTypeVersion()) {
      if ($action == SyncIntent::ACTION_UPDATE) {
        $action = SyncIntent::ACTION_CREATE;
      }
      elseif ($action == SyncIntent::ACTION_DELETE) {
        $version = $this->entity_status->getEntityTypeVersion();
      }
    }

    if ($action == SyncIntent::ACTION_CREATE) {
      $shared_entity_id = NULL;
    }
    elseif ($entity instanceof ConfigEntityInterface) {
      $shared_entity_id = $entity->id();
    }
    else {
      $shared_entity_id = $entity->uuid();
    }

    $storage = $this->getPool()->getCustomStorage($entity_type, $entity_bundle);

    try {
      if ($action === SyncIntent::ACTION_CREATE) {
        $query = $storage->createItem($body);
      }
      elseif ($action === SyncIntent::ACTION_UPDATE) {
        $query = $storage->updateItem($shared_entity_id, $body);
      }
      elseif ($action === SyncIntent::ACTION_DELETE) {
        $query = $storage->deleteItem($shared_entity_id);
      }
      else {
        throw new SyncException(SyncException::CODE_INTERNAL_ERROR, new \Exception("Invalid action $action."));
      }

      $query
        ->setAsDependency(
          $this->flow->getEntityTypeConfig($entity_type, $entity_bundle)['export'] == PushIntent::PUSH_AS_DEPENDENCY
        )
        ->execute();
    }
    catch (NotFoundException $e) {
      // The Sync Core doesn't know this entity yet. This may happen when the Sync Core URL has changed
      // or after a reset of the Sync Core database. In this case we simply try a CREATE instead.
      if ($action === SyncIntent::ACTION_UPDATE) {
        $storage
          ->createItem($body)
          ->setAsDependency(
            $this->flow->getEntityTypeConfig($entity_type, $entity_bundle)['export'] == PushIntent::PUSH_AS_DEPENDENCY
          )
          ->execute();
      }
    }
    catch (SyncCoreException $e) {
      \Drupal::logger('cms_content_sync')->error(
        'Failed to @action entity @entity_type-@entity_bundle @entity_uuid' . PHP_EOL . '@message' . PHP_EOL . 'Got status code @status_code @reason_phrase with body:' . PHP_EOL . '@body',
        [
          '@action' => $action,
          '@entity_type' => $entity_type,
          '@entity_bundle' => $entity_bundle,
          '@entity_uuid' => $entity_uuid,
          '@message' => $e->getMessage(),
          '@status_code' => $e->getStatusCode(),
          '@reason_phrase' => $e->getReasonPhrase(),
          '@body' => $e->getBody() . '',
        ]
      );

      $this->saveFailedPush(PushIntent::PUSH_FAILED_REQUEST_FAILED, $e->getMessage());

      throw new SyncException(SyncException::CODE_PUSH_REQUEST_FAILED, $e);
    }

    if (!$this->entity_status->getLastPush() && !$this->entity_status->getLastPull() && isset($body['url'])) {
      $this->entity_status->set('source_url', $body['url']);
    }
    $this->entity_status->setLastPush($push);

    if ($action == SyncIntent::ACTION_DELETE) {
      $this->entity_status->isDeleted(TRUE);
      $this->pool->markDeleted($entity_type, $entity_uuid);
    }

    if ($version != $this->entity_status->getEntityTypeVersion()) {
      $this->entity_status->setEntityTypeVersion($version);
    }

    $this->entity_status->save();

    // Dispatch entity push event to give other modules the possibility to react on it.
    \Drupal::service('event_dispatcher')->dispatch(AfterEntityPush::EVENT_NAME, new AfterEntityPush($entity, $this->pool, $this->flow, $this->reason, $this->action));

    return TRUE;
  }

  /**
   * @var array
   *   A list of all pushed entities to make sure entities aren't pushed
   *   multiple times during the same request in the format
   *   [$action][$entity_type][$bundle][$uuid] => TRUE
   */
  static protected $pushed;

  /**
   * Check whether the given entity is currently being pushed. Useful to check
   * against hierarchical references as for nodes and menu items for example.
   *
   * @param string $entity_type
   *   The entity type to check for.
   * @param string $uuid
   *   The UUID of the entity in question.
   * @param string $pool
   *   The pool to push to.
   * @param null|string $action
   *   See ::ACTION_*.
   *
   * @return bool
   */
  public static function isPushing($entity_type, $uuid, $pool = NULL, $action = NULL) {
    foreach (self::$pushed as $do => $types) {
      if ($action ? $do != $action : $do == SyncIntent::ACTION_DELETE) {
        continue;
      }
      if (!isset($types[$entity_type])) {
        continue;
      }
      foreach ($types[$entity_type] as $bundle => $entities) {
        if (empty($pool)) {
          if (!empty($entities[$uuid])) {
            return TRUE;
          }
        }
        else {
          if (!empty($entities[$uuid][$pool])) {
            return TRUE;
          }
        }
      }
    }

    return FALSE;
  }

  /**
   * Helper function to push an entity and throw errors if anything fails.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to push.
   * @param string $reason
   *   {@see Flow::PUSH_*}.
   * @param string $action
   *   {@see ::ACTION_*}.
   * @param \Drupal\cms_content_sync\Entity\Flow $flow
   *   The flow to be used. If none is given, all flows that may push this
   *   entity will be asked to do so for all relevant pools.
   * @param \Drupal\cms_content_sync\Entity\Pool $pool
   *   The pool to be used. If not set, all relevant pools for the flow will be
   *   used one after another.
   *
   * @param bool $return_only
   *
   * @return bool|string Whether the entity is configured to be pushed or not.
   *   if $return_only is given, this will return the serialized entity to embed
   *   or NULL.
   *
   * @throws \Drupal\cms_content_sync\Exception\SyncException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \GuzzleHttp\Exception\GuzzleException
   */
  public static function pushEntity(EntityInterface $entity, $reason, $action, Flow $flow = NULL, Pool $pool = NULL, $return_only = FALSE) {
    if (!$flow) {
      $flows = self::getFlowsForEntity($entity, $reason, $action);
      if (!count($flows)) {
        return FALSE;
      }

      $result = FALSE;
      foreach ($flows as $flow) {
        if ($return_only) {
          $result = self::pushEntity($entity, $reason, $action, $flow, NULL, TRUE);
          if ($result) {
            return $result;
          }
        }
        else {
          $result |= self::pushEntity($entity, $reason, $action, $flow);
        }
      }
      return $result;
    }

    if (!$pool) {
      $pools = $flow->getPoolsToPushTo($entity, $reason, $action, TRUE);
      $result = FALSE;
      foreach ($pools as $pool) {
        $infos = EntityStatus::getInfosForEntity($entity->getEntityTypeId(), $entity->uuid(), ['pool' => $pool->label()]);
        $cancel = FALSE;
        foreach ($infos as $info) {
          if (!$info->getFlow()) {
            continue;
          }

          if (!$info->getLastPull()) {
            continue;
          }

          $config = $info->getFlow()->getEntityTypeConfig($entity->getEntityTypeId(), $entity->bundle())['import_updates'];

          if (in_array($config, [PullIntent::PULL_UPDATE_FORCE_AND_FORBID_EDITING, PullIntent::PULL_UPDATE_FORCE_UNLESS_OVERRIDDEN])) {
            $cancel = TRUE;
            break;
          }
        }

        if ($cancel) {
          continue;
        }

        if ($return_only) {
          $result = self::pushEntity($entity, $reason, $action, $flow, $pool, TRUE);
          if ($result) {
            return $result;
          }
        }
        else {
          $result |= self::pushEntity($entity, $reason, $action, $flow, $pool);
        }
      }
      return $result;
    }

    $intent = new PushIntent($flow, $pool, $reason, $action, $entity);
    return $intent->execute($return_only);
  }

  /**
   * @var string NO_PUSH_REASON__JUST_PULLED The entity has been pulled
   *   during this very request, so it can't be pushed again immediately.
   */
  const NO_PUSH_REASON__JUST_PULLED = 'JUST_IMPORTED';

  /**
   * @var string NO_PUSH_REASON__NEVER_PUSHED The entity has never been
   *   pushed before, so pushing the deletion doesn't make sense (it will
   *   not even exist remotely yet).
   */
  const NO_PUSH_REASON__NEVER_PUSHED = 'NEVER_EXPORTED';

  /**
   * @var string NO_PUSH_REASON__UNCHANGED The entity hasn't changed, so the
   *   push would not do anything.
   */
  const NO_PUSH_REASON__UNCHANGED = 'UNCHANGED';

  /**
   * @var string NO_PUSH_REASON__HANDLER_IGNORES The handler for the entity
   *   refused to push this entity. These are usually handler specific
   *   configurations like "Don't push unpublished content" for nodes.
   */
  const NO_PUSH_REASON__HANDLER_IGNORES = 'HANDLER_IGNORES';

  /**
   * @var string NO_PUSH_REASON__NO_POOL No pool was assigned, so there's no push to take place.
   */
  const NO_PUSH_REASON__NO_POOL = 'NO_POOL';

  /**
   * @var array
   *   pushed. Can be queried via self::getNoPushReason($entity). Structure:
   *   [ entity_type_id:string ][ entity_uuid:string ] => string|Exception
   */
  protected static $noPushReasons = [];

  /**
   * Get the reason why a push has not happened.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   * @param bool $as_message
   *
   * @return string|\Exception|null See self::$noPushReasons.
   */
  public static function getNoPushReason($entity, $as_message = FALSE) {
    // If push wasn't even tried, no pool has been assigned.
    if (empty(self::$noPushReasons[$entity->getEntityTypeId()][$entity->uuid()])) {
      $issue = self::NO_PUSH_REASON__NO_POOL;
    }
    else {
      $issue = self::$noPushReasons[$entity->getEntityTypeId()][$entity->uuid()];
    }

    if ($as_message) {
      return self::displayNoPushReason(
        $issue
      );
    }

    return $issue;
  }

  /**
   * Get a user message on why the push failed.
   *
   * @param string|\Exception $reason
   *   The reason from self::getNoPushReason().
   *
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup|string
   */
  public static function displayNoPushReason($reason) {
    if ($reason instanceof \Exception) {
      return $reason->getMessage();
    }

    switch ($reason) {
      case self::NO_PUSH_REASON__HANDLER_IGNORES:
        return \t('The configuration forbids the push.');

      case self::NO_PUSH_REASON__JUST_PULLED:
        return \t('The entity has just been pulled and cannot be pushed immediately with the same request.');

      case self::NO_PUSH_REASON__NEVER_PUSHED:
        return \t('The entity has not been pushed before, so pushing the deletion doesn\'t have any effect.');

      case self::NO_PUSH_REASON__UNCHANGED:
        return \t('The entity has not changed since it\'s last push.');

      default:
        return \t('The entity doesn\'t have any Pool assigned.');
    }

    return NULL;
  }

  /**
   * Helper function to push an entity and display the user the results. If
   * you want to make changes programmatically, use ::pushEntity() instead.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to push.
   * @param string $reason
   *   {@see Flow::PUSH_*}.
   * @param string $action
   *   {@see ::ACTION_*}.
   * @param \Drupal\cms_content_sync\Entity\Flow $flow
   *   The flow to be used. If none is given, all flows that may push this
   *   entity will be asked to do so for all relevant pools.
   * @param \Drupal\cms_content_sync\Entity\Pool $pool
   *   The pool to be used. If not set, all relevant pools for the flow will be
   *   used one after another.
   *
   * @return bool Whether the entity is configured to be pushed or not.
   */
  public static function pushEntityFromUi(EntityInterface $entity, $reason, $action, Flow $flow = NULL, Pool $pool = NULL) {
    $messenger = \Drupal::messenger();
    try {
      $status = self::pushEntity($entity, $reason, $action, $flow, $pool);

      if ($status) {
        if ($action == SyncIntent::ACTION_DELETE) {
          $messenger->addMessage(t('%label has been pushed to your @repository.', ['@repository' => _cms_content_sync_get_repository_name(), '%label' => $entity->getEntityTypeId()]));
        }
        else {
          $messenger->addMessage(t('%label has been pushed to your @repository.', ['@repository' => _cms_content_sync_get_repository_name(), '%label' => $entity->label()]));
        }
        return TRUE;
      }

      return FALSE;
    }
    catch (SyncException $e) {
      $message = $e->parentException ? $e->parentException->getMessage() : (
        $e->errorCode == $e->getMessage() ? '' : $e->getMessage()
      );
      if ($message) {
        $messenger->addWarning(t('Failed to push %label to your @repository (%code). Message: %message', [
          '@repository' => _cms_content_sync_get_repository_name(),
          '%label' => $entity->label(),
          '%code' => $e->errorCode,
          '%message' => $message,
        ]));
      }
      else {
        $messenger->addWarning(t('Failed to push %label to your @repository (%code).', [
          '@repository' => _cms_content_sync_get_repository_name(),
          '%label' => $entity->label(),
          '%code' => $e->errorCode,
        ]));
      }
      self::$noPushReasons[$entity->getEntityTypeId()][$entity->uuid()] = $e;
      return TRUE;
    }
  }

}
