<?php

namespace Drupal\cms_content_sync\SyncCore;

use Drupal\cms_content_sync\SyncCore\Exception\BadRequestException;
use Drupal\cms_content_sync\SyncCore\Exception\ForbiddenException;
use Drupal\cms_content_sync\SyncCore\Exception\NotFoundException;
use Drupal\cms_content_sync\SyncCore\Exception\SyncCoreException;
use Drupal\cms_content_sync\SyncCore\Query\PingQuery;
use Drupal\cms_content_sync\SyncCore\Query\SimpleQuery;
use Drupal\cms_content_sync\SyncCore\Storage\ApiStorage;
use Drupal\cms_content_sync\SyncCore\Storage\ConnectionStorage;
use Drupal\cms_content_sync\SyncCore\Storage\ConnectionSynchronizationStorage;
use Drupal\cms_content_sync\SyncCore\Storage\CustomStorage;
use Drupal\cms_content_sync\SyncCore\Storage\EntityTypeStorage;
use Drupal\cms_content_sync\SyncCore\Storage\InstanceStorage;
use Drupal\cms_content_sync\SyncCore\Storage\MetaInformationConnectionStorage;
use Drupal\cms_content_sync\SyncCore\Storage\ObjectStorage;
use Drupal\cms_content_sync\SyncCore\Storage\PreviewEntityStorage;
use Drupal\cms_content_sync\SyncCore\Storage\RemoteStorageStorage;
use GuzzleHttp\Exception\GuzzleException;

/**
 * Class Client
 * The client used by the Storage to connect to the Sync Core. You can imagine
 * this to be a remote database connection where the Storage talks to individual
 * tables / collections.
 *
 * @package Drupal\cms_content_sync\SyncCore
 */
class Client {
  /**
   * @var string METHOD_GET Perform an HTTP GET request.
   */
  const METHOD_GET = 'GET';
  /**
   * @var string METHOD_POST Perform an HTTP POST request.
   */
  const METHOD_POST = 'POST';
  /**
   * @var string METHOD_PUT Perform an HTTP PUT request.
   */
  const METHOD_PUT = 'PUT';
  /**
   * @var string METHOD_PATCH Perform an HTTP PATCH request.
   */
  const METHOD_PATCH = 'PATCH';
  /**
   * @var string METHOD_DELETE Perform an HTTP DELETE request.
   */
  const METHOD_DELETE = 'DELETE';

  /**
   * @var int MAX_ITEMS_PER_PAGE
   *    The maximum number of items that can be queried per page as restricted by the Sync Core.
   */
  const MAX_ITEMS_PER_PAGE = 100;

  /**
   * @var string
   *    The base URL of the remote Sync Core. See Pool::$backend_url
   */
  protected $base_url;

  /**
   * @var \GuzzleHttp\Client
   *    The HTTP client used to execute requests.
   */
  protected $client;

  /**
   * @var ApplicationInterface
   */
  protected $application;

  /**
   * @param ApplicationInterface $application
   * @param string $base_url
   */
  public function __construct($application, $base_url) {
    $this->application = $application;
    $this->base_url = $base_url;

    $this->client = $application->getHttpClient();
  }

  /**
   * @return string
   */
  public function getBaseUrl() {
    return $this->base_url;
  }

  const LOG_LEVEL_ERROR = 'error';

  const LOG_LEVEL_WARNING = 'warn';

  const STATUS_PATH = '/status';

  const LOG_PATH = '/log';

  /**
   * @param string $api_id
   * @param string $site_id
   * @param string $entity_type_name
   * @param string $bundle_name
   *
   * @return \Drupal\cms_content_sync\SyncCore\Storage\CustomStorage
   */
  public function getCustomStorage($api_id, $site_id, $entity_type_name, $bundle_name) {
    static $cache;

    if (!empty($cache[$api_id][$site_id][$entity_type_name][$bundle_name])) {
      return $cache[$api_id][$site_id][$entity_type_name][$bundle_name];
    }

    return $cache[$api_id][$site_id][$entity_type_name][$bundle_name] = new CustomStorage(
      $this,
      $api_id,
      $site_id,
      $entity_type_name,
      $bundle_name
    );
  }

  /**
   * @param string[]|string|null $level
   * @param bool $obfuscateCredentials
   *
   * @return null|array
   *
   * @throws Exception\SyncCoreException
   */
  public function getLog($level = NULL, $obfuscateCredentials = TRUE) {
    $arguments = [];
    if ($level) {
      $arguments['level'] = (is_array($level) ? implode(',', $level) : $level);
    }

    $items = SimpleQuery
      ::create($this, self::LOG_PATH, $arguments)
        ->execute()
        ->getResult()['items'];

    if ($obfuscateCredentials) {
      $items = Helper::obfuscateCredentials($items);
    }

    foreach ($items as &$item) {
      if (isset($item['err']['message'])) {
        $item['msg'] = $item['err']['message'];
      }
    }

    return $items;
  }

  /**
   * @return array
   *
   * @throws Exception\SyncCoreException
   */
  public function getStatus() {
    return SimpleQuery
      ::create($this, self::STATUS_PATH)
        ->execute()
        ->getResult();
  }

  /**
   * @param $site_url
   * @param $method
   * @param $authentication
   *
   * @return bool
   *
   * @throws \Drupal\cms_content_sync\SyncCore\Exception\SyncCoreException
   */
  public function requestPing($site_url, $method, $authentication) {
    return PingQuery
      ::create($this)
        ->setSiteUrl($site_url)
        ->setMethod($method)
        ->setAuthentication($authentication)
        ->execute()
        ->succeeded();
  }

  /**
   * Execute a GET request.
   *
   * @param \Drupal\cms_content_sync\SyncCore\Query\Query $query
   *
   * @return mixed
   *
   * @throws \Drupal\cms_content_sync\SyncCore\Exception\SyncCoreException
   */
  public function request($query) {
    $params = http_build_query($query->toArray());
    $url = $this->base_url . $query->getPath() . ($params ? '?' . $params : '');

    $method = $query->getMethod();
    $body = $query->getBody();

    try {
      $options = [
        'http_errors' => FALSE,
      ];

      if ($body !== NULL) {
        $options['headers']['Content-Type'] = 'application/json';
        $options['body'] = json_encode($body);
      }

      if ($method === Client::METHOD_DELETE) {
        $options['timeout'] = 15;
      }

      $response = $this->client->request(
        $method,
        $url,
        $options
      );
    }
    catch (\Exception $e) {
      throw new SyncCoreException($e->getMessage());
    }
    catch (GuzzleException $e) {
      throw new SyncCoreException($e->getMessage());
    }

    $status = $response->getStatusCode();

    $returnBoolean = $query->returnBoolean();
    if ($returnBoolean && $status !== 200 && $status !== 201) {
      return FALSE;
    }

    if ($status === 400) {
      throw new BadRequestException('The Sync Core responded with 400 Bad Request.', $status, $response->getReasonPhrase(), $response->getBody());
    }
    elseif ($status === 403) {
      throw new ForbiddenException('The Sync Core responded with 403 Forbidden.', $status, $response->getReasonPhrase(), $response->getBody());
    }
    elseif ($status === 404) {
      throw new NotFoundException('The Sync Core responded with 404 Not Found.', $status, $response->getReasonPhrase(), $response->getBody());
    }
    elseif ($status !== 200 && $status !== 201) {
      throw new SyncCoreException('The Sync Core responded with a non-OK status code.', $status, $response->getReasonPhrase(), $response->getBody());
    }

    $data = json_decode($response->getBody(), TRUE);

    if ($returnBoolean) {
      if (is_string($returnBoolean)) {
        return $data[$returnBoolean];
      }

      return TRUE;
    }

    return $data;
  }

  /**
   * Receive an absolute URL to the Sync Core and provide a relative URL in return.
   * If you have a '_resource_url' somewhere, use this function to convert it to a relative URL you can then pass to
   * a simple query for example.
   *
   * @param string $url
   *
   * @return string
   */
  public static function getRelativeUrl($url) {
    if (substr($url, 0, 7) == 'http://' || substr($url, 0, 8) == 'https://') {
      // Remove protocol, host and path prefix so the path is relative and can be inserted into the normal Query classes.
      return preg_replace('@^https?://[^/]+/rest@', '', $url);
    }

    // Is already relative.
    return $url;
  }

  /**
   * @param string $type
   *
   * @return \Drupal\cms_content_sync\SyncCore\Storage\Storage
   *
   * @throws \Exception
   */
  public function getStorageById($type) {
    static $cache;

    if ($type == MetaInformationConnectionStorage::ID) {
      return $this->getMetaInformationConnectionStorage();
    }
    if ($type == ConnectionStorage::ID) {
      return $this->getConnectionStorage();
    }
    if ($type == EntityTypeStorage::ID) {
      return $this->getEntityTypeStorage();
    }
    if ($type == ConnectionSynchronizationStorage::ID) {
      return $this->getConnectionSynchronizationStorage();
    }
    if ($type == PreviewEntityStorage::ID) {
      return $this->getPreviewEntityStorage();
    }

    if (isset($cache[$type])) {
      return $cache[$type];
    }

    if ($type == ObjectStorage::ID) {
      return $cache[$type] = new ObjectStorage($this);
    }
    if ($type == RemoteStorageStorage::ID) {
      return $cache[$type] = new RemoteStorageStorage($this);
    }
    if ($type == InstanceStorage::ID) {
      return $cache[$type] = new InstanceStorage($this);
    }
    if ($type == ApiStorage::ID) {
      return $cache[$type] = new ApiStorage($this);
    }

    throw new \Exception("Unknown storage type $type.");
  }

  /**
   * @return \Drupal\cms_content_sync\SyncCore\Storage\ConnectionSynchronizationStorage
   */
  public function getConnectionSynchronizationStorage() {
    static $cache;
    if (!empty($cache)) {
      return $cache;
    }
    return $cache = new ConnectionSynchronizationStorage($this);
  }

  /**
   * @return \Drupal\cms_content_sync\SyncCore\Storage\ConnectionStorage
   */
  public function getConnectionStorage() {
    static $cache;
    if (!empty($cache)) {
      return $cache;
    }
    return $cache = new ConnectionStorage($this);
  }

  /**
   * @return \Drupal\cms_content_sync\SyncCore\Storage\InstanceStorage
   */
  public function getInstanceStorage() {
    static $cache;
    if (!empty($cache)) {
      return $cache;
    }
    return $cache = new InstanceStorage($this);
  }

  /**
   * @return \Drupal\cms_content_sync\SyncCore\Storage\MetaInformationConnectionStorage
   */
  public function getMetaInformationConnectionStorage() {
    static $cache;
    if (!empty($cache)) {
      return $cache;
    }
    return $cache = new MetaInformationConnectionStorage($this);
  }

  /**
   * @return \Drupal\cms_content_sync\SyncCore\Storage\EntityTypeStorage
   */
  public function getEntityTypeStorage() {
    static $cache;
    if (!empty($cache)) {
      return $cache;
    }
    return $cache = new EntityTypeStorage($this);
  }

  /**
   * @return \Drupal\cms_content_sync\SyncCore\Storage\PreviewEntityStorage
   */
  public function getPreviewEntityStorage() {
    static $cache;
    if (!empty($cache)) {
      return $cache;
    }
    return $cache = new PreviewEntityStorage($this);
  }

}
