<?php

namespace Drupal\cms_content_sync\SyncCore\Storage;

use Drupal\cms_content_sync\SyncCore\Query\CreateItemQuery;
use Drupal\cms_content_sync\SyncCore\Query\DeleteItemQuery;
use Drupal\cms_content_sync\SyncCore\Query\ItemQuery;
use Drupal\cms_content_sync\SyncCore\Query\ListQuery;
use Drupal\cms_content_sync\SyncCore\Query\UpdateItemQuery;

/**
 * Class Storage
 * A remote entity type storage for the Sync Core. Provides Query objects for
 * the given pool and the given entity type. One Storage class is available per
 * Entity Type.
 *
 * @package Drupal\cms_content_sync\SyncCore\Storage
 */
abstract class Storage {
  /**
   * ID property.
   */
  const PROPERTY_ID = 'id';

  /**
   * @var \Drupal\cms_content_sync\SyncCore\Client
   */
  protected $client;

  /**
   * Storage constructor.
   *
   * @param \Drupal\cms_content_sync\SyncCore\Client $client
   */
  public function __construct($client) {
    $this->client = $client;
  }

  /**
   * Get the entity type ID.
   *
   * @return string
   */
  abstract public function getId();

  /**
   * Get the path to append to the Pool::$backend_url to query entities of this
   * type.
   *
   * @return string
   */
  public function getPath() {
    return '/' . $this->getId();
  }

  /**
   * Get the client that this Storage belongs to.
   *
   * @return \Drupal\cms_content_sync\SyncCore\Client
   */
  public function getClient() {
    return $this->client;
  }

  /**
   * Create and return an instance of the ListQuery.
   *
   * @return \Drupal\cms_content_sync\SyncCore\Query\ListQuery
   */
  public function listItems() {
    return new ListQuery($this);
  }

  /**
   * Create and return an instance of the ItemQuery.
   *
   * @param string $id
   *
   * @return \Drupal\cms_content_sync\SyncCore\Query\ItemQuery
   */
  public function getItem($id) {
    $query = new ItemQuery($this);
    $query->setEntityId($id);
    return $query;
  }

  /**
   * Create and return an instance of the CreateItemQuery.
   *
   * @param array $item
   *
   * @return \Drupal\cms_content_sync\SyncCore\Query\CreateItemQuery
   */
  public function createItem($item) {
    $query = new CreateItemQuery($this);
    $query->setItem($item);
    return $query;
  }

  /**
   * Create and return an instance of the UpdateItemQuery.
   *
   * @param string $id
   * @param array $item
   *
   * @return \Drupal\cms_content_sync\SyncCore\Query\UpdateItemQuery
   */
  public function updateItem($id, $item) {
    $query = new UpdateItemQuery($this);
    $query->setEntityId($id);
    $query->setItem($item);
    return $query;
  }

  /**
   * Create and return an instance of the DeleteItemQuery.
   *
   * @param string $id
   *
   * @return \Drupal\cms_content_sync\SyncCore\Query\DeleteItemQuery
   */
  public function deleteItem($id) {
    $query = new DeleteItemQuery($this);
    $query->setEntityId($id);
    return $query;
  }

}
