<?php

namespace Drupal\cms_content_sync;

use Drupal\cms_content_sync\Controller\AuthenticationByUser;
use Drupal\cms_content_sync\Controller\ContentSyncSettings;
use Drupal\cms_content_sync\Entity\Flow;
use Drupal\cms_content_sync\Entity\Pool;
use Drupal\cms_content_sync\Event\BeforeEntityTypeExport;
use Drupal\cms_content_sync\Plugin\rest\resource\EntityResource;
use Drupal\cms_content_sync\Plugin\Type\EntityHandlerPluginManager;
use Drupal\cms_content_sync\SyncCore\Exception\SyncCoreException;
use Drupal\cms_content_sync\SyncCore\Query\Condition\DataCondition;
use Drupal\cms_content_sync\SyncCore\Query\Condition\ParentCondition;
use Drupal\cms_content_sync\SyncCore\Storage\ApiStorage;
use Drupal\cms_content_sync\SyncCore\Storage\ConnectionStorage;
use Drupal\cms_content_sync\SyncCore\Storage\ConnectionSynchronizationStorage;
use Drupal\cms_content_sync\SyncCore\Storage\CustomStorage;
use Drupal\cms_content_sync\SyncCore\Storage\EntityTypeStorage;
use Drupal\cms_content_sync\SyncCore\Storage\InstanceStorage;
use Drupal\cms_content_sync\SyncCore\Storage\ObjectStorage;
use Drupal\cms_content_sync\SyncCore\Storage\PreviewEntityStorage;
use Drupal\cms_content_sync\SyncCore\Storage\RemoteStorageStorage;
use Drupal\Core\Serialization\Yaml;
use Drupal\cms_content_sync\Form\PoolForm;

/**
 * Class SyncCoreFlowExport used to export the Synchronization config to the Sync
 * Core backend.
 */
class SyncCoreFlowExport extends SyncCoreExport {

  /**
   * The type to identify what kind of object we want to query in the general object storage.
   */
  const OBJECT_STORAGE_TYPE = 'content_sync-drupal-flow-config';
  const OBJECT_STORAGE_PROPERTY_MODULE_VERSION = ObjectStorage::PROPERTY_PROPERTIES . '.module_version';
  const OBJECT_STORAGE_PROPERTY_POOLS = ObjectStorage::PROPERTY_PROPERTIES . '.pools';

  /**
   * @var \Drupal\cms_content_sync\Entity\Flow
   */
  protected $flow;

  /**
   * Sync Core Config constructor.
   *
   * @param \Drupal\cms_content_sync\Entity\Flow $flow
   *   The flow this exporter is used for.
   */
  public function __construct(Flow $flow) {
    $pools = $flow->getUsedPools();
    $first = reset($pools);
    parent::__construct($first->getClient());

    $this->flow = $flow;
  }

  /**
   * Get a list of all Sync Core connections as resource URLs.
   *
   * @return \Drupal\cms_content_sync\SyncCore\Entity\Connection[]
   */
  protected function getConnectionEntities() {
    $pools = Pool::getAll();
    $result = [];

    foreach ($this->flow->getEntityTypeConfig() as $id => $type) {
      $entity_type_name = $type['entity_type_name'];
      $bundle_name = $type['bundle_name'];

      if ($type['handler'] == Flow::HANDLER_IGNORE) {
        continue;
      }

      $entity_type_pools = [];
      if (isset($type['import_pools'])) {
        foreach ($type['import_pools'] as $pool_id => $state) {
          if (!isset($entity_type_pools[$pool_id])) {
            $entity_type_pools[$pool_id] = [];
          }
          $entity_type_pools[$pool_id]['import'] = $state;
        }
      }

      if (isset($type['export_pools'])) {
        foreach ($type['export_pools'] as $pool_id => $state) {
          if (!isset($entity_type_pools[$pool_id])) {
            $entity_type_pools[$pool_id] = [];
          }
          $entity_type_pools[$pool_id]['export'] = $state;
        }
      }

      foreach ($entity_type_pools as $pool_id => $definition) {
        $pool = $pools[$pool_id];
        $export = $definition['export'] ?? NULL;
        $import = $definition['import'] ?? NULL;

        if ($export == Pool::POOL_USAGE_FORBID && $import == Pool::POOL_USAGE_FORBID) {
          continue;
        }

        $url     = $pool->getBackendUrl();
        $api     = $pool->id;
        $site_id = $pool->getSiteId();

        $result[] = $pool
          ->getClient()
          ->getConnectionStorage()
          ->getEntity(CustomStorage::getCustomId($api, $site_id, $entity_type_name, $bundle_name));
      }
    }

    return $result;
  }

  /**
   * Get a list of all Sync Core connection synchronizations as resource URLs.
   *
   * @param array $filters
   *   Additional filters for the synchronization URLs to return.
   *   - entity_type_id: Only return synchronizations for this entity type.
   *    - bundle:         Only return synchronizations for this bundle.
   *    - flow_export:    Only return synchronizations with this export setting.
   *    - flow_import:    Only return synchronizations with this import setting.
   *    - export_pools:   Only return synchronizations with this pool export setting.
   *    - import_pools:   Only return synchronizations with this pool import setting.
   *
   * @return \Drupal\cms_content_sync\SyncCore\Entity\ConnectionSynchronization[]
   */
  protected function getConnectionSynchronizationEntities($filters) {
    $pools = Pool::getAll();
    $result = [];

    foreach ($this->flow->getEntityTypeConfig() as $id => $type) {
      $entity_type_name = $type['entity_type_name'];
      if (!empty($filters['entity_type_id']) && $filters['entity_type_id'] != $entity_type_name) {
        continue;
      }

      $bundle_name = $type['bundle_name'];
      if (!empty($filters['bundle']) && $filters['bundle'] != $bundle_name) {
        continue;
      }

      if ($type['handler'] == Flow::HANDLER_IGNORE) {
        continue;
      }

      if (!empty($filters['flow_export']) && !in_array($type['export'], $filters['flow_export'])) {
        continue;
      }

      if (!empty($filters['flow_import']) && !in_array($type['import'], $filters['flow_import'])) {
        continue;
      }

      $entity_type_pools = [];
      if (isset($type['import_pools'])) {
        foreach ($type['import_pools'] as $pool_id => $state) {
          if (!isset($entity_type_pools[$pool_id])) {
            $entity_type_pools[$pool_id] = [];
          }
          $entity_type_pools[$pool_id]['import'] = $state;
        }
      }
      if (isset($type['export_pools'])) {
        foreach ($type['export_pools'] as $pool_id => $state) {
          if (!isset($entity_type_pools[$pool_id])) {
            $entity_type_pools[$pool_id] = [];
          }
          $entity_type_pools[$pool_id]['export'] = $state;
        }
      }

      foreach ($entity_type_pools as $pool_id => $definition) {
        if (empty($pools[$pool_id])) {
          continue;
        }
        $pool = $pools[$pool_id];
        $export = $definition['export'] ?? NULL;
        $import = $definition['import'] ?? NULL;

        if (!empty($filters['pool_import']) && !in_array($import, $filters['pool_import'])) {
          continue;
        }

        if (!empty($filters['pool_export']) && !in_array($export, $filters['pool_export'])) {
          continue;
        }

        $api     = $pool->id;
        $site_id = $pool->getSiteId();

        $local_connection_id = CustomStorage::getCustomId($api, $site_id, $entity_type_name, $bundle_name);
        $sync_id = ConnectionSynchronizationStorage::getExternalConnectionSynchronizationId($local_connection_id, !empty($filters['flow_export']) && !empty($filters['pool_export']));
        $result[$sync_id] = $pool
          ->getClient()
          ->getConnectionSynchronizationStorage()
          ->getEntity($sync_id);
      }
    }

    return $result;
  }

  /**
   * Kindly ask the Sync Core to pull an individual entity, e.g. when the
   * override checkbox is removed.
   *
   * @param string $entity_type_id
   *   The entity type.
   * @param string $bundle
   *   The bundle.
   * @param string $shared_entity_id
   *   The entity UUID.
   *
   * @return array
   */
  public function pull($entity_type_id, $bundle, $shared_entity_id) {
    /**
     * @var \Drupal\cms_content_sync\SyncCore\Entity\ConnectionSynchronization[] $synchronizations
     */
    $synchronizations = array_merge($this->getConnectionSynchronizationEntities(
      [
        'flow_import'     => [PullIntent::PULL_AUTOMATICALLY, PullIntent::PULL_AS_DEPENDENCY],
        'pool_import'     => [Pool::POOL_USAGE_FORCE],
        'entity_type_id'  => $entity_type_id,
        'bundle'          => $bundle,
      ]
    ), $this->getConnectionSynchronizationEntities(
      [
        'flow_import'     => [PullIntent::PULL_AS_DEPENDENCY, PullIntent::PULL_MANUALLY],
        'pool_import'     => [Pool::POOL_USAGE_ALLOW],
        'entity_type_id'  => $entity_type_id,
        'bundle'          => $bundle,
      ]
    ));

    $result = [];

    $type = $this->flow->getEntityTypeConfig($entity_type_id, $bundle);
    $manual = $type['import'] == PullIntent::PULL_MANUALLY;

    foreach ($synchronizations as $id => $synchronization) {
      try {
        $result[$id] = $synchronization
          ->synchronizeSingle($shared_entity_id)
          ->isManual($manual)
          ->execute()
          ->succeeded();
      }
      catch (\Exception $e) {
        $result[$id] = FALSE;
      }
    }

    return $result;
  }

  /**
   * Kindly ask the Sync Core to pull all entities that are auto pulled or pushed.
   *
   * @param bool $export
   *   Whether to push (TRUE) or pull (FALSE).
   *
   * @param bool $force
   *
   * @return array An associative array with the synchronization URLs as keys.
   *   The value is FALSE for failed requests. Otherwise it's an associative
   *   array with the following keys:
   *   - id:     ID for this process. If none is given, no entities had to be synchronized.
   *   - total:  Total number of entities to be synchronized, if any.
   */
  public function startSync($export, $force = FALSE) {
    $synchronizations = $this->getConnectionSynchronizationEntities(
      [
        'flow_export' => $export ? [PushIntent::PUSH_FORCED] : NULL,
        'pool_export' => $export ? [Pool::POOL_USAGE_FORCE, Pool::POOL_USAGE_ALLOW] : NULL,
        'flow_import' => !$export ? ($force ? [PullIntent::PULL_AUTOMATICALLY] : [PullIntent::PULL_AUTOMATICALLY, PullIntent::PULL_MANUALLY]) : NULL,
        'pool_import' => !$export ? ($force ? [Pool::POOL_USAGE_FORCE] : [Pool::POOL_USAGE_FORCE, Pool::POOL_USAGE_ALLOW]) : NULL,
      ]
    );

    $result = [];

    foreach ($synchronizations as $id => $synchronization) {
      try {
        $result[$id] = $synchronization
          ->synchronizeAll()
          ->force($force)
          ->execute()
          ->getResult();
      }
      catch (\Exception $e) {
        $result[$id] = FALSE;
      }
    }

    return $result;
  }

  /**
   * Kindly ask the Sync Core to login to this site.
   */
  public function login() {
    $connections = $this->getConnectionEntities();

    $result = [];

    foreach ($connections as $connection) {
      try {
        $result[] = $connection
          ->login()
          ->execute()
          ->succeeded();
      }
      catch (SyncCoreException $e) {
        $result[] = FALSE;
      }
    }

    return $result;
  }

  /**
   * Create all entity types, connections and synchronizations as required.
   *
   * TODO: Provide static function to export all Flows at once as right now overlapping Pools usage between Flows will
   *       Result in multiple CREATE requests for the same configuration, slowing the export down a little.
   *
   * @throws \Exception If the user profile for pull is not available.
   */
  public function prepareBatch() {
    $remote_storages = [];

    $operations = new SyncCoreBatchCollection();

    $this->addStorages($remote_storages, $operations);

    foreach (Flow::getAll() as $id => $flow) {
      if ($this->flow->id === $id) {
        continue;
      }

      $sub = new SyncCoreFlowExport($flow);
      $sub->addStorages($remote_storages, $operations, TRUE);
    }

    foreach ($remote_storages as $pool_id => $storage) {
      $operations->add($storage, RemoteStorageStorage::ID);
    }

    return $operations->getOperations();
  }

  /**
   * Create all entity types, connections and synchronizations as required.
   *
   * @param $remote_storages
   * @param SyncCoreBatchCollection $operations
   * @param bool $extend_only
   *
   * @throws \Exception If the user profile for pull is not available.
   */
  public function addStorages(&$remote_storages, &$operations, $extend_only = FALSE) {
    // Ignore disabled flows at export.
    if (!$this->flow->get('status')) {
      return;
    }

    $export_url = ContentSyncSettings::getInstance()->getSiteBaseUrl();
    $enable_preview = ContentSyncSettings::getInstance()->isPreviewEnabled();

    $cms_content_sync_disable_optimization = boolval(\Drupal::config('cms_content_sync.debug')
      ->get('cms_content_sync_disable_optimization'));

    $authentication_provider = AuthenticationByUser::getInstance();

    $entity_types = $this->flow->sync_entities;

    $pools = Pool::getAll();

    $this->remove(TRUE);

    $export_pools = [];

    foreach ($this->flow->getEntityTypeConfig() as $id => $type) {
      $entity_type_name = $type['entity_type_name'];
      $bundle_name      = $type['bundle_name'];
      $version          = $type['version'];

      if ($type['handler'] == Flow::HANDLER_IGNORE) {
        continue;
      }

      $current = Flow::getEntityTypeVersion($entity_type_name, $bundle_name);
      if ($current !== $version) {
        throw new \Exception("Entity type $entity_type_name.$bundle_name was changed without updating Flow {$this->flow->id}. Please re-save that Flow first to apply the latest entity type changes.");
      }

      $handler = $this->flow->getEntityTypeHandler($type);

      $entity_type_pools = [];
      if (isset($type['import_pools'])) {
        foreach ($type['import_pools'] as $pool_id => $state) {
          if (!isset($entity_type_pools[$pool_id])) {
            $entity_type_pools[$pool_id] = [];
          }

          if ($type['import'] == PullIntent::PULL_DISABLED) {
            $entity_type_pools[$pool_id]['import'] = Pool::POOL_USAGE_FORBID;
            continue;
          }

          $entity_type_pools[$pool_id]['import'] = $state;
        }
      }

      if (isset($type['export_pools'])) {
        foreach ($type['export_pools'] as $pool_id => $state) {
          if (!isset($entity_type_pools[$pool_id])) {
            $entity_type_pools[$pool_id] = [];
          }

          if ($type['export'] == PushIntent::PUSH_DISABLED) {
            $entity_type_pools[$pool_id]['export'] = Pool::POOL_USAGE_FORBID;
            continue;
          }

          $entity_type_pools[$pool_id]['export'] = $state;
        }
      }

      foreach ($entity_type_pools as $pool_id => $definition) {
        if (empty($pools[$pool_id])) {
          continue;
        }

        $pool = $pools[$pool_id];

        $export = NULL;
        $import = NULL;

        if (isset($definition['export'])) {
          $export = $definition['export'];
        }

        if (isset($definition['import'])) {
          $import = $definition['import'];
        }

        if ((!$export || $export == Pool::POOL_USAGE_FORBID) && (!$import || $import == Pool::POOL_USAGE_FORBID)) {
          continue;
        }

        if (!in_array($pool, $export_pools)) {
          $export_pools[] = $pool;
        }

        $api     = $pool->id;
        $site_id = $pool->getSiteId();

        if ($extend_only && !isset($remote_storages[$api])) {
          continue;
        }

        if (strlen($site_id) > PoolForm::siteIdMaxLength) {
          throw new \Exception(t('The site id of pool ' . $pool_id . ' is having more then ' . PoolForm::siteIdMaxLength . ' characters. This is not allowed due to backend limitations and will result in an exception when it is trying to be exported.'));
        }

        if ($pool->getAuthenticationType() == Pool::AUTHENTICATION_TYPE_BASIC_AUTH) {
          $authentication = [
            'type' => 'basic_auth',
            'username' => $authentication_provider->getUsername(),
            'password' => $authentication_provider->getPassword(),
            'base_url' => $export_url,
          ];
        }
        else {
          $authentication = [
            'type' => 'drupal8_services',
            'username' => $authentication_provider->getUsername(),
            'password' => $authentication_provider->getPassword(),
            'base_url' => $export_url,
          ];
        }

        $entity_type_id = EntityTypeStorage::getExternalEntityTypeId($api, $entity_type_name, $bundle_name, $version);

        $local_connection_id = CustomStorage::getCustomId($api, $site_id, $entity_type_name, $bundle_name);

        $pull_condition = [];

        if (EntityHandlerPluginManager::isEntityTypeFieldable($entity_type_name)) {
          $entityFieldManager = \Drupal::service('entity_field.manager');
          /** @var \Drupal\Core\Field\FieldDefinitionInterface[] $fields */
          $fields = $entityFieldManager->getFieldDefinitions($entity_type_name, $bundle_name);

          $forbidden = $handler->getForbiddenFields();

          foreach ($fields as $key => $field) {
            if (!isset($entity_types[$id . '-' . $key])) {
              continue;
            }

            if (in_array($key, $forbidden)) {
              continue;
            }

            if (isset($entity_type['new_properties'][$key])) {
              continue;
            }

            if (!empty($entity_types[$id . '-' . $key]['handler_settings']['subscribe_only_to'])) {
              $allowed = [];

              foreach ($entity_types[$id . '-' . $key]['handler_settings']['subscribe_only_to'] as $ref) {
                $allowed[] = $ref['uuid'];
              }

              $pull_condition[] = DataCondition::in($key . '.*.' . SyncIntent::UUID_KEY, $allowed);
            }
          }
        }

        if (count($pull_condition) > 1) {
          $pull_condition = ParentCondition::all($pull_condition);
        }
        elseif (count($pull_condition)) {
          $pull_condition = $pull_condition[0];
        }
        else {
          $pull_condition = NULL;
        }

        if ($extend_only) {
          if ($operations->get($entity_type_id, EntityTypeStorage::ID)) {
            // If this Flow pushes or pulls AUTOMATICALLY, overwrite the AS_DEPENDENCY setting of the original Flow.
            // because the Sync Core will only know about one or the other. So we make it the AUTOMATICALLY and then filter
            // ourselves when the request comes in.
            if ($extend_only) {
              if ($import != Pool::POOL_USAGE_FORBID && $type['import'] != PullIntent::PULL_DISABLED) {
                $sync_id = ConnectionSynchronizationStorage::getExternalConnectionSynchronizationId($local_connection_id, FALSE);
                $sync = $operations->get($sync_id, ConnectionSynchronizationStorage::ID);
                if ($sync) {
                  // Add another mode (see below)
                  $sync['options']['modes'][] = [
                    'id' => $this->flow->id,
                    'is_manual' => $type['import'] == PullIntent::PULL_MANUALLY,
                    'is_dependent' => $type['import'] == PullIntent::PULL_AS_DEPENDENCY,
                    'condition' => $pull_condition ? $pull_condition->toArray() : NULL,
                    'delete' => boolval($type['import_deletion_settings']['import_deletion']),
                  ];

                  $operations->add($sync, ConnectionSynchronizationStorage::ID);
                }
              }
            }

            continue;
          }
        }

        $entity_type = [
          'id' => $entity_type_id,
          'name_space' => $entity_type_name,
          'name' => $bundle_name,
          'version' => $version,
          'base_class' => "api-unify/services/drupal/v0.1/models/base.model",
          'custom' => TRUE,
          'new_properties' => [
            'source' => [
              'type' => 'reference',
              'default_value' => NULL,
              'connection_identifiers' => [
                [
                  'properties' => [
                    'id' => 'source_connection_id',
                  ],
                ],
              ],
              'model_identifiers' => [
                [
                  'properties' => [
                    'id' => 'source_id',
                  ],
                ],
              ],
              'multiple' => FALSE,
            ],
            'source_id' => [
              'type' => 'id',
              'default_value' => NULL,
            ],
            'source_connection_id' => [
              'type' => 'id',
              'default_value' => NULL,
            ],
            'preview' => [
              'type' => 'string',
              'default_value' => NULL,
            ],
            'url' => [
              'type' => 'string',
              'default_value' => NULL,
            ],
            'apiu_translation' => [
              'type' => 'object',
              'default_value' => NULL,
            ],
            'metadata' => [
              'type' => 'object',
              'default_value' => NULL,
            ],
            'embed_entities' => [
              'type' => 'object',
              'default_value' => NULL,
              'multiple' => TRUE,
            ],
            'menu_items' => [
              'type' => 'object',
              'default_value' => NULL,
              'multiple' => TRUE,
            ],
            'title' => [
              'type' => 'string',
              'default_value' => NULL,
            ],
            'created' => [
              'type' => 'int',
              'default_value' => NULL,
            ],
            'changed' => [
              'type' => 'int',
              'default_value' => NULL,
            ],
            'uuid' => [
              'type' => 'string',
              'default_value' => NULL,
            ],
          ],
          'new_property_lists' => [
            'list' => [
              '_resource_url' => 'value',
              '_resource_connection_id' => 'value',
              'id' => 'value',
            ],
            'reference' => [
              '_resource_url' => 'value',
              '_resource_connection_id' => 'value',
              'id' => 'value',
            ],
            'details' => [
              '_resource_url' => 'value',
              '_resource_connection_id' => 'value',
              'id' => 'value',
              'source' => 'reference',
              'apiu_translation' => 'value',
              'metadata' => 'value',
              'embed_entities' => 'value',
              'title' => 'value',
              'created' => 'value',
              'changed' => 'value',
              'uuid' => 'value',
              'url' => 'value',
              'menu_items' => 'value',
            ],
            'database' => [
              'id' => 'value',
              'source_id' => 'value',
              'source_connection_id' => 'value',
              'preview' => 'value',
              'url' => 'value',
              'apiu_translation' => 'value',
              'metadata' => 'value',
              'embed_entities' => 'value',
              'title' => 'value',
              'created' => 'value',
              'changed' => 'value',
              'uuid' => 'value',
              'menu_items' => 'value',
            ],
            'modifiable' => [
              'title' => 'value',
              'preview' => 'value',
              'url' => 'value',
              'apiu_translation' => 'value',
              'metadata' => 'value',
              'embed_entities' => 'value',
              'menu_items' => 'value',
            ],
            'required' => [
              'uuid' => 'value',
            ],
          ],
          'api_id' => $api . '-' . ApiStorage::CUSTOM_API_VERSION,
        ];

        if (EntityHandlerPluginManager::isEntityTypeFieldable($entity_type_name)) {

          $entityFieldManager = \Drupal::service('entity_field.manager');
          /** @var \Drupal\Core\Field\FieldDefinitionInterface[] $fields */
          $fields = $entityFieldManager->getFieldDefinitions($entity_type_name, $bundle_name);

          $forbidden = $handler->getForbiddenFields();

          foreach ($fields as $key => $field) {
            if (!isset($entity_types[$id . '-' . $key])) {
              continue;
            }

            if (in_array($key, $forbidden)) {
              continue;
            }

            if (isset($entity_type['new_properties'][$key])) {
              continue;
            }

            $entity_type['new_properties'][$key] = [
              'type' => 'object',
              'default_value' => NULL,
              'multiple' => TRUE,
            ];

            $entity_type['new_property_lists']['details'][$key] = 'value';
            $entity_type['new_property_lists']['database'][$key] = 'value';

            if ($field->isRequired()) {
              $entity_type['new_property_lists']['required'][$key] = 'value';
            }

            if (!$field->isReadOnly() || $key == 'menu_link' || $field->getType() == 'path' || ($entity_type_name == 'crop' && ($key == 'x' || $key == 'y'))) {
              $entity_type['new_property_lists']['modifiable'][$key] = 'value';
            }
          }
        }

        // TODO entity types should also contain the entity type handler in their machine name, preventing the following potential errors:
        // - Different flows may use different entity type handlers, resulting in different entity type definitions for the same entity type
        // - Changing the entity type handler must change the entity type definition which will not work if we don't update the machine name.
        // => Alternatively, force the same handler to be used at all sites.
        $handler->updateEntityTypeDefinition($entity_type);

        // Dispatch EntityTypeExport event to give other modules the possibility
        // to adjust the entity type definition and add custom fields.
        \Drupal::service('event_dispatcher')->dispatch(BeforeEntityTypeExport::EVENT_NAME, new BeforeEntityTypeExport($entity_type_name, $bundle_name, $entity_type));

        // Create the entity type.
        $operations->add($entity_type, EntityTypeStorage::ID);

        $pool_connection_id = CustomStorage::getCustomId($api, InstanceStorage::POOL_SITE_ID, $entity_type_name, $bundle_name);
        // Create the pool connection entity for this entity type.
        $operations->add([
          'id' => $pool_connection_id,
          'name' => 'Drupal pool connection for ' . $entity_type_name . '-' . $bundle_name . '-' . $version,
          'hash' => CustomStorage::getCustomPath($api, InstanceStorage::POOL_SITE_ID, $entity_type_name, $bundle_name),
          'usage' => 'EXTERNAL',
          'status' => 'READY',
          'options' => [
            'update_all' => $cms_content_sync_disable_optimization,
          ],
          'entity_type_id' => $entity_type_id,
        ], ConnectionStorage::ID);

        // Create a synchronization from the pool to the preview connection.
        if ($enable_preview) {
          $operations->add([
            'id' => $pool_connection_id . '--to--preview',
            'name' => 'Synchronization Pool ' . $entity_type_name . '-' . $bundle_name . ' -> Preview',
            'options' => [
              'create_entities' => TRUE,
              'update_entities' => TRUE,
              'delete_entities' => TRUE,
              'update_none_when_loading' => TRUE,
              'exclude_reference_properties' => [
                'pSource',
              ],
            ],
            'status' => 'READY',
            'source_connection_id' => $pool_connection_id,
            'destination_connection_id' => PreviewEntityStorage::ID,
          ], ConnectionSynchronizationStorage::ID);
        }

        if (isset($remote_storages[$api])) {
          $entity_type_id_without_version = EntityTypeStorage::getExternalEntityTypeId($api, $entity_type_name, $bundle_name, '');
          $skip = FALSE;

          // If the given entity type has added before, we don't add it a second time.
          foreach ($remote_storages[$api]['entity_type_ids'] as $id) {
            if (substr($id, 0, strlen($entity_type_id_without_version)) === $entity_type_id_without_version) {
              $skip = TRUE;
            }
          }

          if (!$skip) {
            $remote_storages[$api]['entity_type_ids'][] = $entity_type_id;
          }
        }
        else {
          $remote_storages[$api] = [
            'id' => RemoteStorageStorage::getStorageId($api, $site_id),
            'name' => 'Drupal connection on ' . $site_id . ' for ' . $api,
            'status' => 'READY',
            'instance_id' => $site_id,
            'api_id' => $api,
            'entity_type_ids' => [$entity_type_id],
            'connection_id_pattern' => CustomStorage::getCustomId('[api.id]', '[instance.id]', '[entity_type.name_space]', '[entity_type.name]'),
            'connection_name_pattern' => 'Drupal connection on [instance.id] for [entity_type.name_space].[entity_type.name]:[entity_type.version]',
            'connection_path_pattern' => CustomStorage::getCustomPath('[api.id]', '[instance.id]', '[entity_type.name_space]', '[entity_type.name]'),
            'connection_options' => [
              'list_url' => EntityResource::getInternalUrl('[api.id]', '[entity_type.name_space]', '[entity_type.name]', '[entity_type.version]'),
              'item_url' => EntityResource::getInternalUrl('[api.id]', '[entity_type.name_space]', '[entity_type.name]', '[entity_type.version]', '[id]'),
              'authentication' => $authentication,
              'update_all' => $cms_content_sync_disable_optimization,
            ],
          ];
        }

        if (!$extend_only) {
          // Create a synchronization from the pool to the local connection.
          if ($import != Pool::POOL_USAGE_FORBID && $type['import'] != PullIntent::PULL_DISABLED) {
            $operations->add([
              'id' => ConnectionSynchronizationStorage::getExternalConnectionSynchronizationId($local_connection_id, FALSE),
              'name' => 'Synchronization for ' . $entity_type_name . '/' . $bundle_name . '/' . $version . ' from Pool -> ' . $site_id,
              'options' => [
                'dependency_connection_id' => CustomStorage::DEPENDENCY_CONNECTION_ID,
                'modes' => [
                  [
                    'id' => $this->flow->id,
                    'is_manual' => $type['import'] == PullIntent::PULL_MANUALLY,
                    'is_dependent' => $type['import'] == PullIntent::PULL_AS_DEPENDENCY,
                    'condition' => $pull_condition ? $pull_condition->toArray() : NULL,
                    'delete' => boolval($type['import_deletion_settings']['import_deletion']),
                  ],
                ],
                'force_updates' => $cms_content_sync_disable_optimization,
                'create_entities' => TRUE,
                'update_entities' => TRUE,
                'delete_entities' => TRUE,
                'update_none_when_loading' => TRUE,
                'exclude_reference_properties' => [
                  'pSource',
                ],
              ],
              'status' => 'READY',
              'source_connection_id' => $pool_connection_id,
              'destination_connection_id' => $local_connection_id,
            ], ConnectionSynchronizationStorage::ID);
          }
          if ($export != Pool::POOL_USAGE_FORBID && $type['export'] != PushIntent::PUSH_DISABLED) {
            $operations->add([
              'id' => ConnectionSynchronizationStorage::getExternalConnectionSynchronizationId($local_connection_id, TRUE),
              'name' => 'Synchronization for ' . $entity_type_name . '/' . $bundle_name . '/' . $version . ' from ' . $site_id . ' -> Pool',
              'options' => [
                'dependency_connection_id' => CustomStorage::POOL_DEPENDENCY_CONNECTION_ID,
                // As entities will only be sent to Sync Core if the sync config
                // allows it, the synchronization entity doesn't need to filter
                // any further.
                'create_entities' => TRUE,
                'update_entities' => TRUE,
                'delete_entities' => TRUE,
                'force_updates' => $cms_content_sync_disable_optimization,
                'dependent_entities_only' => FALSE,
                'update_none_when_loading' => TRUE,
                'exclude_reference_properties' => [
                  'pSource',
                ],
              ],
              'status' => 'READY',
              'source_connection_id' => $local_connection_id,
              'destination_connection_id' => $pool_connection_id,
            ], ConnectionSynchronizationStorage::ID);
          }
        }
      }
    }

    if ($extend_only) {
      return;
    }

    // Store the flow in the simple object storage of the Sync Core. This allows other sites to query for existing
    // configuration to clone or mirror. Used by the `CopyRemoteFlow` form.
    $pool_ids = [];
    $url = NULL;
    $site_id = NULL;
    foreach ($export_pools as $pool) {
      $pool_ids[$pool->id()] = TRUE;
      $site_id = $pool->getSiteId();
    }
    $module_info = \Drupal::service('extension.list.module')->getExtensionInfo('cms_content_sync');

    $operations->add([
      'id' => self::OBJECT_STORAGE_TYPE . '-' . $site_id . '-' . $this->flow->id(),
      'type' => self::OBJECT_STORAGE_TYPE,
      'properties' => [
        'module_version' => $module_info['version'] ? $module_info['version'] : 'dev',
        'pools' => $pool_ids,
        'id' => $this->flow->id(),
        'name' => $this->flow->label(),
        'site' => $site_id,
        'config' => Yaml::encode(\Drupal::service('config.storage')->read('cms_content_sync.flow.' . $this->flow->id())),
      ],
    ], ObjectStorage::ID);

    // Always export required pools as well to prevent any potential issues
    // TODO: Optimize when called via Drush to not send the same pool requests for multiple Flows.
    $pool_operations = [];
    foreach ($export_pools as $pool) {
      $exporter = new SyncCorePoolExport($pool);

      $pool_operations = array_merge(
        $pool_operations,
        $exporter->prepareBatch(!!count($pool_operations))
      );
    }

    $operations->addMultiple($pool_operations, TRUE);
  }

  /**
   * Delete the synchronizations from this connection.
   */
  public function remove($removedOnly = TRUE) {
    // @TODO Refactor for pool changes
    return TRUE;
  }

}
