<?php

namespace Drupal\cms_content_sync;

use Drupal\cms_content_sync\Controller\ContentSyncSettings;
use Drupal\cms_content_sync\Entity\Pool;
use Drupal\cms_content_sync\SyncCore\Query\Condition\DataCondition;
use Drupal\cms_content_sync\SyncCore\Query\Condition\ParentCondition;
use Drupal\cms_content_sync\SyncCore\Storage\ApiStorage;
use Drupal\cms_content_sync\SyncCore\Storage\ConnectionStorage;
use Drupal\cms_content_sync\SyncCore\Storage\InstanceStorage;
use Drupal\cms_content_sync\SyncCore\Storage\PreviewEntityStorage;

/**
 *
 */
class SyncCorePoolExport extends SyncCoreExport {

  /**
   * @var \Drupal\cms_content_sync\Entity\Pool
   */
  protected $pool;

  /**
   * SyncCorePoolExport constructor.
   *
   * @param \Drupal\cms_content_sync\Entity\Pool $pool
   *   The pool this exporter is used for.
   */
  public function __construct(Pool $pool) {
    parent::__construct($pool->getClient());

    $this->pool = $pool;
  }

  /**
   * Check if another site has already connected to this pool with the same site ID.
   *
   * @return bool
   *
   * @throws SyncCore\Exception\SyncCoreException
   */
  public function siteIdExists() {
    $condition = ParentCondition::all()
      ->add(DataCondition::equal(InstanceStorage::PROPERTY_ID, $this->pool->getSiteId()))
      ->add(DataCondition::notEqual(InstanceStorage::PROPERTY_BASE_URL, ContentSyncSettings::getInstance()->getSiteBaseUrl()))
      ->add(DataCondition::notEqual(InstanceStorage::PROPERTY_BASE_URL, NULL));

    return $this
      ->client
      ->getInstanceStorage()
      ->listItems()
      ->setCondition(
        $condition
      )
      ->execute()
      ->getNumberOfItems() > 0;
  }

  /**
   * @inheritdoc
   */
  public function prepareBatch($subsequent = FALSE) {
    $base_url = ContentSyncSettings::getInstance()->getSiteBaseUrl();
    if (empty($base_url)) {
      throw new \Exception('Please provide a base_url via settings or drush command.');
    }

    $this->remove(TRUE);

    $operations = new SyncCoreBatchCollection();

    // Skip creation of base entities if they are already created.
    if (!$subsequent) {
      // Create "drupal" API entity.
      $operations->add([
        'id' => 'drupal-' . ApiStorage::CUSTOM_API_VERSION,
        'name' => 'drupal',
        'version' => ApiStorage::CUSTOM_API_VERSION,
      ], ApiStorage::ID);

      // Create the preview connection entity.
      $operations->add([
        'id' => PreviewEntityStorage::ID,
        'name' => 'Drupal preview connection',
        'hash' => PreviewEntityStorage::EXTERNAL_PREVIEW_PATH,
        'usage' => 'EXTERNAL',
        'status' => 'READY',
        'entity_type_id' => PreviewEntityStorage::PREVIEW_ENTITY_ID,
        'options' => [
          'crud' => [
            'read_list' => [],
          ],
          'static_values' => [],
          PreviewEntityStorage::PUBLIC_ACCESS_OPTION_NAME => ContentSyncSettings::getInstance()->isDirectSyncCoreAccessEnabled(),
        ],
      ], ConnectionStorage::ID);
    }

    // Create the child entity.
    $operations->add([
      'id' => $this->pool->id . '-' . ApiStorage::CUSTOM_API_VERSION,
      'name' => $this->pool->label(),
      'version' => ApiStorage::CUSTOM_API_VERSION,
      'parent_id' => 'drupal-' . ApiStorage::CUSTOM_API_VERSION,
    ], ApiStorage::ID);

    // Create the instance entity.
    $operations->add([
      'id' => $this->pool->getSiteId(),
      'base_url' => $base_url,
      'version' => SyncCore::get()->getApplicationModuleVersion(),
      'api_id' => $this->pool->id . '-' . ApiStorage::CUSTOM_API_VERSION,
    ], InstanceStorage::ID);

    return $operations->getOperations();
  }

  /**
   *
   */
  public function remove($removedOnly = TRUE) {
    return TRUE;
  }

}
